# -*- coding: utf-8 -*-
"""Parent class for QueueHandleAdapterReader and QueuePipeAdapterWriter"""
from __future__ import unicode_literals

import io
import random
import tempfile  # For comparisons

from pickle import PicklingError  # nosec
from threading import Thread  # For non-multi-processing queues
from typing import Union

# Multiprocessing imports
import multiprocessing

from _io import _IOBase  # For comparisons

from .classtemplate import ClassTemplate
from .exceptionhandler import HandleAlreadySet
from .common import DIRECTION
from .common import is_threaded
from .common import UNION_SUPPORTED_EVENTS, UNION_SUPPORTED_LOCKS, UNION_SUPPORTED_QUEUES


class MessageCounter(object):
    """Track the number of messages processed by an Adapter.

    Access the count with MessageCenter.value"""
    def __init__(self,
                 start_method: str=None):
        # Which multiprocess context to use
        self.multiprocessing_ctx = multiprocessing.get_context(start_method)

        self.counter = self.multiprocessing_ctx.Value('Q')
        self.counter.value = 0

    def increment(self) -> None:
        """Increment the message counter."""
        self.counter.value += 1

    def __getattr__(self, attr):
        if attr == 'value':
            return object.__getattribute__(self, 'counter').value

        return object.__getattribute__(self, attr)


class _QueueHandleAdapterBase(ClassTemplate):
    """QueueHandleAdapter abstract implementation"""
    def __init__(self,
                 queue: UNION_SUPPORTED_QUEUES,
                 *,  # End of positional arguments
                 subclass_name: str,
                 queue_direction: DIRECTION,
                 name: str=None,
                 handle: io.IOBase=None,
                 log_name: str=None,
                 start_method: str=None,
                 thread_only: bool=False,
                 **kwargs):
        """
        Args:
            queue: The queue to attach
            subclass_name: Name of the concrete class implementation
            queue_direction: Indicate direction relative to pipe;
                e.g. from stdout, the flow is (PIPE => QUEUE)
                => CLIENT QUEUES (from pipe/queue into client queues), and
                therefore queue_direction would be "source".
            name: Name to use in logging for this instance
            handle: The IO handle
            log_name: Additional name to use in logging for this instance
            start_method: Name of the multiprocessing start method to use (fork, forkserver, spawn)
            thread_only: True to keep processing in a single thread using threading
        """
        # Unique ID for this PrPipe
        # Not used for cryptographic purposes, so excluding from Bandit
        self.id = \
            ''.join([random.choice(  # nosec
                '0123456789ABCDEF') for x in range(6)])

        # Name for this instance, typically stdin/stdout/stderr
        self.name = name

        # Name of the subclass (QueueHandleAdapterReader, QueuePipeAdapterWriter)
        self.subclass_name = subclass_name

        # Additional naming for differentiating multiple instances
        self.log_name = log_name

        # Initialize the logger
        self._initialize_logging_with_log_name(self.subclass_name)

        # Which "direction" client queues will use in the queue_link
        self.queue_direction = queue_direction
        self.client_direction = DIRECTION.FROM if queue_direction == DIRECTION.TO \
            else DIRECTION.TO

        # Whether to use only threading (vs multiprocessing)
        self.thread_only = thread_only

        # Which multiprocess context to use
        self.multiprocessing_ctx = multiprocessing.get_context(start_method)

        # Whether we have ever been started
        self.started = self.multiprocessing_ctx.Event()

        # Whether we have been asked to stop
        self.stopped = self.multiprocessing_ctx.Event()

        # Whether the queue adapter process should stop
        self.stop_event = self.multiprocessing_ctx.Event()

        # The queue proxy to be used as the main input or output buffer.
        # Attaching this to the queue link is the responsibility of subclasses.
        self.queue = queue

        # Lock to notify readers/writers that a read/write is in progress
        self.queue_lock = self.multiprocessing_ctx.Lock()

        # Store the number of messages processed
        # Q unsigned long long https://docs.python.org/3/library/array.html#module-array
        self.messages_processed = MessageCounter(start_method=start_method)

        # Store other args
        self.kwargs = kwargs

        self.process = None
        self.handle = None
        if handle is not None:
            self.set_handle(handle)

    # Class contains Locks and Queues which cannot be pickled
    def __getstate__(self):
        """Prevent _QueueHandleAdapterBase from being pickled across Processes

        Raises:
            Exception
        """
        raise PicklingError("Don't pickle me!")

    def set_handle(self, handle: Union[io.IOBase, str]) -> None:
        """Set the handle to read from or write to.

        Args:
            handle: An open handle (subclasses of file, IO.IOBase)

        Raises:
            HandleAlreadySet
        """
        if self.process is not None:
            raise HandleAlreadySet

        # Store the pipehandle
        self.handle = handle

        # Process name
        process_name = f'{self.subclass_name}-{self.name}'

        if self.log_name is not None:
            process_name = f'{process_name}-{self.log_name}'

        # If the handle is a BufferedReader we have to stay in the same process
        if isinstance(handle, (_IOBase, tempfile._TemporaryFileWrapper)) and not self.thread_only:
            self._log.warning('Adapter process %s will only be threaded as we cannot '
                              'send an open handle to another process.', process_name)
            self.thread_only = True

        # Select the right concurrency mechanism
        threaded = is_threaded(self.queue)
        # pylint: disable=invalid-name
        Parallel = Thread if threaded or self.thread_only else self.multiprocessing_ctx.Process
        # pylint: enable=invalid-name

        # Arguments for
        arg_dict = {
            "name": self.name,
            "handle": handle,
            "queue": self.queue,
            "queue_lock": self.queue_lock,
            "stop_event": self.stop_event,
            "messages_processed": self.messages_processed}
        arg_dict.update(self.kwargs)

        self._log.debug("Setting %s adapter process for %s pipe handle",
                        self.subclass_name, self.name)
        self.process = Parallel(target=self.queue_handle_adapter,
                                name=process_name,
                                kwargs=arg_dict)
        self.process.daemon = True
        self.process.start()
        self.started.set()
        self._log.debug("Kicked off %s adapter process for %s pipe handle",
                        self.subclass_name, self.name)

    def __del__(self):
        self.close()

    @staticmethod
    def queue_handle_adapter(*,  # All named parameters are required keyword arguments
                             name: str,
                             handle: io.IOBase,
                             queue: UNION_SUPPORTED_QUEUES,
                             queue_lock: UNION_SUPPORTED_LOCKS,
                             stop_event: UNION_SUPPORTED_EVENTS,
                             messages_processed: MessageCounter,
                             trusted: bool,
                             **kwargs) -> None:
        """Override me in a subclass to do something useful"""

    def _stop(self) -> None:
        """Internal stop method"""
        # Mark that we have been asked to stop
        self.stopped.set()

        # Stop the adapter
        self.stop_event.set()

        while True:
            self.process.join(timeout=1)

            if self.process.is_alive():
                pass
            else:
                break

    def close(self) -> None:
        """Stop the adapter and queue link and clean up.

        Does not force a drain of the queues.
        """
        if hasattr(self, '_stop') and hasattr(self, 'started'):
            if self.started:
                self._stop()

        # Delete/unlink other resources
        attributes = [
            'started',
            'stopped',
            'stop_event',
            'queue_lock',
            'queue',
            'handle',
            'messages_processed']

        for attr in attributes:
            if hasattr(self, attr):
                setattr(self, attr, None)

    def is_alive(self) -> bool:
        """Check whether the thread/process managing the movement is still active

        Returns:
            True if the adapter thread/process is still running, False otherwise.
        """
        return self.process.is_alive()

    def is_drained(self) -> bool:
        """Check alive and empty

        Attempts clean semantic response to "is there, or will there be, data
        to read?"

        Returns:
            True if fully drained, False if not
        """
        drained = True

        # If we aren't started, we have to stop here. The process isn't ready
        # to call is_alive()
        if not self.started.is_set():
            return False

        # Alive (True) means we are not drained
        drained = drained and not self.is_alive()

        return drained

    def get_messages_processed(self) -> int:
        """Return the number of messages moved by this adapter

        Returns:
            The number of messages processed by this adapter.
        """
        return self.messages_processed.value



"""Store and retrieve results from throughput tests"""
import os
import sqlite3
import sys

from datetime import datetime
from typing import Union

DB_NAME = 'throughput/throughput.sqlite.db'

CURRENT_SESSION_TABLE_NAME = 'current_session'
CURRENT_SESSION_TABLE_SCHEMA = 'parent_process_id, session_id, session_time, ' \
                               'UNIQUE(parent_process_id)'

TEST_SESSION_TABLE_NAME = 'sessions'
TEST_SESSION_TABLE_SCHEMA = 'session_id, time, UNIQUE(session_id)'

RESULT_TABLE_NAME = 'results'
RESULT_TABLE_SCHEMA = 'session_id, python_version, test_name, start_method, source, destination, ' \
                      'result, result_unit'


class ThroughputResults(object):
    """Manage results storage for QueueLink throughput tests."""
    def __init__(self,
                 session_id: str,
                 session_time: datetime,
                 start_method: str,
                 source_path: str,
                 dest_path: str):
        self.session_id = session_id
        self.session_time = session_time
        self.start_method = start_method
        self.source_path = source_path
        self.dest_path = dest_path
        self.db = sqlite3.connect(DB_NAME)

        # Create tables if needed
        cursor = self.db.cursor()
        for tbl, schema in [(TEST_SESSION_TABLE_NAME, TEST_SESSION_TABLE_SCHEMA),
                            (RESULT_TABLE_NAME, RESULT_TABLE_SCHEMA),
                            (CURRENT_SESSION_TABLE_NAME, CURRENT_SESSION_TABLE_SCHEMA)]:
            result = cursor.execute('SELECT name FROM sqlite_master WHERE name=?', (tbl,))
            if result.fetchone() is None:
                cursor.execute('CREATE TABLE IF NOT EXISTS ?(?)', (tbl, schema))

        # Session entry
        self.update_session_id()
        self.start_session()

    def update_session_id(self):
        """Uses parent PID to determine if we should use a different session ID and start time"""
        ppid = os.getppid()
        cursor = self.db.cursor()
        result = cursor.execute('SELECT parent_process_id, session_id, session_time FROM '
                                f'{CURRENT_SESSION_TABLE_NAME} '  # nosec
                                'WHERE parent_process_id = ?',
                                (ppid,))
        result_tuple = result.fetchone()

        if result_tuple is None:
            # None found, empty the table
            cursor.execute('DELETE FROM '
                           f'{CURRENT_SESSION_TABLE_NAME}')  # nosec

            time = self.session_time.strftime('%Y-%m-%d %H:%M:%S')
            cursor.execute('INSERT INTO '
                           f'{CURRENT_SESSION_TABLE_NAME} '  # nosec
                           'VALUES ?, ?, ?',
                           (ppid, self.session_id, time))
            self.db.commit()

        else:
            session_id = result_tuple[1]
            session_time = result_tuple[2]

            self.session_id = session_id
            self.session_time = datetime.strptime(session_time, '%Y-%m-%d %H:%M:%S')

    def start_session(self):
        """Establish start time and write session information to DB."""
        time = self.session_time.strftime('%Y-%m-%d %H:%M:%S')
        cursor = self.db.cursor()
        cursor.execute('INSERT OR IGNORE INTO '
                       f'{TEST_SESSION_TABLE_NAME} '  # nosec
                       'VALUES (?, ?)',
                       (self.session_id, time))
        self.db.commit()

    def put(self, test_name: str,
            result: Union[int, float],
            result_unit: str):
        """Store a test result"""
        python_version = f'{sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}'
        cursor = self.db.cursor()
        cursor.execute('INSERT INTO '
                       f'{RESULT_TABLE_NAME} '  # nosec
                       'VALUES (?, ?, ?, ?, ?, ?, ?, ?)',
                       (self.session_id, python_version, test_name, self.start_method,
                        self.source_path, self.dest_path, result, result_unit))
        self.db.commit()


class ThroughputResultsOutput():
    """Format throughput results."""
    def __init__(self):
        self.db = sqlite3.connect(DB_NAME)
        self.db.row_factory = sqlite3.Row  # Set the kind of result objects that get returned

    def get_latest_session_id(self):
        """Get the session ID of the latest throughput session"""
        cursor = self.db.cursor()
        sql = ('SELECT session_id FROM '
               f'{TEST_SESSION_TABLE_NAME} '  # nosec
               'ORDER BY time DESC LIMIT 1')
        results = cursor.execute(sql).fetchone()

        return results['session_id'] if len(results) > 0 else None

    def get_session_results(self, session_id: str=None):
        """Get results of the specified (or latest) session"""
        if session_id is None:
            session_id = self.get_latest_session_id()

        # Records
        sql = ('SELECT python_version, test_name, start_method, source, destination, '
               '  CAST(ROUND(AVG(result), 6) AS TEXT) as result_avg, result_unit '
               f'FROM {RESULT_TABLE_NAME} '  # nosec
               'WHERE session_id = ? '
               'GROUP BY python_version, test_name, start_method, source, destination, '
               '  result_unit '
               'ORDER BY python_version DESC, test_name ASC, start_method ASC, source ASC, '
               '  destination ASC')
        cursor = self.db.cursor()
        results = cursor.execute(sql, (session_id,)).fetchall()

        # Add the column names as the first row
        if len(results) > 0:
            results.insert(0, results[0].keys())

        return results


if __name__ == "__main__":
    # Print the latest set of results
    results_object = ThroughputResultsOutput()
    results_list = results_object.get_session_results()

    print(f'Total rows: {len(results_list)-1}')

    # Column widths
    # https://sqlpey.com/python/top-5-methods-to-create-nicely-formatted-column-outputs-in-python/
    widths = [max(map(len, col))+2 for col in zip(*results_list)]

    # Print rows
    for i, row in enumerate(results_list):
        # if i > 20:
        #     break
        print(" ".join(val.ljust(width) for val, width in zip(row, widths)))



# -*- coding: utf-8 -*-
"""Get the package version"""
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version


def get_version():
    """Get the package version"""
    try:
        _version = version(__name__)
    except PackageNotFoundError:
        # package is not installed
        _version = "0.0.0"

    return _version



"""Measure throughput of QueueLink and the adapters

For each of the concurrency types (threading, fork, forkserver, spawn):
Time to first element
Elements per second
Element transit latency
"""
import itertools
import logging
import multiprocessing
import os
import queue
import statistics

from datetime import datetime
from queue import Empty
from threading import Thread  # For non-multi-processing queues
from typing import Union

from queuelink import QueueLink, DIRECTION, safe_get
from queuelink.timer import Timer
from queuelink.common import PROC_START_METHODS, QUEUE_TYPE_LIST, is_threaded
from queuelink.throughput_results import ThroughputResults

# Queue type list plus start methods
CARTESIAN_QUEUE_TYPES_START_LIST = itertools.product(QUEUE_TYPE_LIST,
                                                  PROC_START_METHODS)


def send_elements(src_q: QUEUE_TYPE_LIST, text: str, element_count: int):
    """Function to add arbitrary amount of data to a queue
    Run in a separate thread/process
    """
    for i in range(element_count):
        src_q.put(text)


class ConcurrentContext(object):
    """Handle concurrency objects"""
    def __init__(self, start_method: str=None):
        self.start_method = start_method
        self.manager = None

        if self.start_method:
            self.multiprocessing_ctx = multiprocessing.get_context(self.start_method)
        else:
            self.multiprocessing_ctx = None

    def start_manager(self):
        """Start a multiprocessing manager if one is in use."""
        if not self.start_method:
            raise AttributeError('Cannot use a manager without a start method')

        if not self.manager:
            self.manager = self.multiprocessing_ctx.Manager()

    def stop(self):
        """Stop a multiprocessing manager if one is in use."""
        if hasattr(self.manager, 'shutdown'):
            self.manager.shutdown()

    def Process(self, *args, **kwargs):  # pylint: disable=invalid-name
        """Create a context-appropriate Process"""
        return self.multiprocessing_ctx.Process(*args, **kwargs)

    def parallel_factory(self, source_q: QUEUE_TYPE_LIST) -> Union[Thread, Process]:
        """Retrieve the appropriate Thread or Process class for parallel execution"""
        threaded = is_threaded(source_q)

        # Decide whether to use a local thread or process-based publisher
        Parallel = Thread if threaded else self.Process

        return Parallel

    def queue_factory(self, module: str, class_name: str):
        """Return a queue from the given module and class."""
        if module == 'queue':
            return getattr(queue, class_name)()

        if module == 'multiprocessing':
            return getattr(self.multiprocessing_ctx, class_name)()

        if module == 'manager':
            self.start_manager()
            return getattr(self.manager, class_name)()


class Throughput(object):
    """Abstract class for other Throughput classes."""
    def __init__(self, start_method: str):
        self.timeout = 60  # Some spawn instances needed a little more time
        self.text = 'a😂' * 10
        self.logger_name = f'queuelink.throughput.{start_method}'
        self.log = logging.getLogger(self.logger_name)

        # Where to find sample commands and test logging config
        content_dir = os.path.join(os.path.dirname(__file__), '..', 'content')
        sample_command_path = os.path.join(content_dir, 'line_output.py')
        self.sample_command_path = sample_command_path

        self.start_method = start_method
        self.ctx = ConcurrentContext(start_method=start_method)

    def stop(self):
        """Stop the internal multiprocessing manager."""
        self.ctx.stop()

    def queue_factory(self, *args, **kwargs):
        return self.ctx.queue_factory(*args, **kwargs)


class Throughput_QueueLink(Throughput):
    def __init__(self, start_method: str, source_type, dest_type,
                 session_id: str=None, session_time: datetime=None):
        super().__init__(start_method=start_method)

        self.source_module = source_type[0]
        self.source_class = source_type[1]
        self.source_path = f'{source_type[0]}.{source_type[1]}'

        self.dest_module = dest_type[0]
        self.dest_class = dest_type[1]
        self.dest_path = f'{dest_type[0]}.{dest_type[1]}'

        # For results reporting
        self.results = ThroughputResults(
            session_id=session_id,
            session_time=session_time,
            start_method=self.start_method,
            source_path=f'{self.source_path}',
            dest_path=f'{self.dest_path}'
        )

    def get_source_q(self):
        return self.queue_factory(module=self.source_module, class_name=self.source_class)

    def get_dest_q(self):
        return self.queue_factory(module=self.dest_module, class_name=self.dest_class)

    def get_from_q(self, target_q):
        return safe_get(queue_obj=target_q, timeout=self.timeout)

    def time_to_first_element(self):
        """Measure how long it takes for the first element to be available"""
        source_q = self.get_source_q()
        dest_q = self.get_dest_q()
        queue_link = QueueLink(name='throughput', source=source_q, start_method=self.start_method)

        # Add to the source queue and register
        source_q.put(self.text)

        # Start the timer and register the destination queue (starting the link)
        start = Timer.now()
        queue_link.register_queue(q=dest_q, direction=DIRECTION.TO)

        # Start trying to get the element
        object_out = self.get_from_q(dest_q)
        end = Timer.now()
        timing = round(end - start, 8)

        self.results.put(test_name='time_to_first_element',
                         result=str(timing),
                         result_unit='seconds')

        self.log.info('Time to first element for source %s and destination %s, start method %s: %s',
                      self.source_path, self.dest_path, self.start_method, timing)

    def avg_time_per_element_after_first_queuelink(self):
        """Measure the nominal latency"""
        iterations = 500
        source_q = self.get_source_q()
        dest_q = self.get_dest_q()
        queue_link = QueueLink(name='avg_throughput', source=source_q, destination=dest_q,
                               start_method=self.start_method)

        # Move the first one through so we only time elements after the link has started
        source_q.put(self.text)
        self.get_from_q(dest_q)

        # Do more
        timing_list = []
        for _ in range(iterations):
            # Place the next item into the source
            source_q.put(self.text)
            start = Timer.now()

            # Start trying to get the element
            object_out = self.get_from_q(dest_q)
            end = Timer.now()
            timing = end - start

            timing_list.append(timing)

        queue_link.close()

        results = {'mean': round(sum(timing_list) / len(timing_list), 8),
                   'median': round(statistics.median(timing_list), 8),
                   'stddev': round(statistics.pstdev(timing_list), 8)}

        self.results.put(test_name='avg_time_per_element_after_first_queuelink',
                         result=results['mean'],
                         result_unit='seconds')

        self.results.put(test_name='median_time_per_element_after_first_queuelink',
                         result=results['median'],
                         result_unit='seconds')

        self.results.put(test_name='stddev_time_per_element_after_first_queuelink',
                         result=results['stddev'],
                         result_unit='seconds')

        self.log.info('Results for time per element across %s elements with source '
                      '%s and destination %s, start method %s: %s',
                      iterations, self.source_path, self.dest_path, self.start_method, results)

    def elements_per_second_queuelink(self):
        """Measure the number of elements per second"""
        test_q = self.get_source_q()
        source_q = self.get_source_q()
        dest_q = self.get_dest_q()
        queue_link = QueueLink(name='avg_throughput',
                               source=source_q,
                               destination=dest_q,
                               start_method=self.start_method)

        def elements_per_second(element_count, src_q, dst_q):

            Parallel = self.ctx.parallel_factory(src_q)
            source_proc = Parallel(
                target=send_elements,
                name='element_source',
                kwargs={
                    'src_q': src_q,
                    'text': self.text,
                    'element_count': element_count})
            source_proc.daemon = True

            # Pull them out
            source_proc.start()
            timer = Timer()
            ts_list = []
            for i in range(element_count):  # Iterate the number of elements
                while True: # Need to get one and one only element per cycle of the for loop
                    try:
                        self.get_from_q(dst_q)
                        ts_list.append(timer.now())
                        break  # Break the while loop

                    except Empty:
                        continue  # Continue the while loop until we have this one element

            # Join/cleanup the source process
            source_proc.join()

            # Start and end
            start = ts_list[0]
            end = ts_list[-1]

            # Check if the test ended in less than 1 second
            if end - start < 1:
                self.log.debug('Elements/second test for source %s and destination '
                               '%s ended with %i in %f s',
                               self.source_path, self.dest_path, element_count, end - start)
                raise ValueError

            # Find which item index crossed the 1 second mark
            for i, timestamp in enumerate(ts_list):
                if timestamp - start >= 1:
                    return i+1

            raise AssertionError('Looks like we did not exceed 1 second in runtime.')

        def elements_per_seconds_increase(src_q, dst_q):
            element_count = 1000

            while True:
                try:
                    self.log.debug('Elements in per second increase with source %s and '
                                   'destination %s: %i',
                                   self.source_path, self.dest_path, element_count)
                    eps = elements_per_second(element_count=element_count, src_q=src_q, dst_q=dst_q)
                    return eps

                except ValueError:
                    element_count *= 10

        # Baseline read rate from a queue
        baseline_eps = elements_per_seconds_increase(src_q=test_q, dst_q=test_q)

        self.results.put(test_name='elements_per_second_queuelink_baseline',
                         result=str(baseline_eps),
                         result_unit='elements_per_second')

        # Actual rate from the link
        actual_eps = elements_per_seconds_increase(src_q=source_q, dst_q=dest_q)

        self.results.put(test_name='elements_per_second_queuelink_actual',
                         result=str(actual_eps),
                         result_unit='elements_per_second')

        self.log.info('Elements per second with baseline (added/removed from one queue) '
                      'on %s: %i, and actual (two queues connected by a QueueLink) based on '
                      'source %s and destination %s: %i',
                      self.source_path, baseline_eps, self.source_path,
                      self.dest_path, actual_eps)



# -*- coding: utf-8 -*-
"""Template class to standardize repetitive functionality"""
# pylint: disable=too-few-public-methods

import logging


class ClassTemplate(object):
    """A pattern to hold boilerplate code across classes"""
    id = None
    log_name = None
    name = None
    _log = None

    def _initialize_logging(self, class_name: str) -> None:
        """Basic logging with class name

        Args:
            class_name: String to use for the class name
        """
        if getattr(self, "_log", None) is not None:
            return

        # Logging
        self._log = logging.getLogger(class_name)
        self.add_logging_handler(logging.NullHandler())

    def _initialize_logging_with_id(self, class_name: str) -> None:
        """Adds `self.id` to logger name after class name

        Args:
            class_name: String to use for the class name
        """
        if getattr(self, '_log', None) is not None:
            return

        # Logger name
        log_name = f'{class_name}-{self.id}'

        # Logging
        self._log = logging.getLogger(log_name)
        self.add_logging_handler(logging.NullHandler())

    def _initialize_logging_with_log_name(self, class_name: str) -> None:
        """Use an extended name when recording log lines

        Will include `class_name`, `self.id`, `self.log_name`, and `self.name`
            if those are populated (in that order).

        Args:
            class_name: String to use for the class name
        """
        if getattr(self, '_log', None) is not None:
            return

        # Logger name
        log_name = f'{class_name}-{self.id}'

        if getattr(self, 'log_name', None) is not None:
            log_name = f'{log_name}.{self.log_name}'

        if getattr(self, 'name', None) is not None:
            log_name = f'{log_name}.{self.name}'

        # Logging
        self._log = logging.getLogger(log_name)
        self.add_logging_handler(logging.NullHandler())

    def add_logging_handler(self, handler: logging.Handler) -> None:
        """Add a logging handler to the logger

        Args:
            handler: A logging handler
        """
        self._log.addHandler(handler)



# -*- coding: utf-8 -*-
"""Factory function for automatically wiring sources and destinations together.

Pass in any supported source and destination and ``link()`` selects and
instantiates the correct combination of ``QueueLink``,
``QueueHandleAdapterReader``, and/or ``QueueHandleAdapterWriter``.
"""
from __future__ import unicode_literals

import multiprocessing
import time

from enum import Enum
from io import IOBase
from os import PathLike

from multiprocessing.managers import BaseProxy

from .common import DIRECTION, THREADED_QUEUES
from .contentwrapper import WRAP_WHEN
from .exceptionhandler import ProcessNotStarted
from .queue_handle_adapter_reader import QueueHandleAdapterReader
from .queue_handle_adapter_writer import QueueHandleAdapterWriter
from .queuelink import QueueLink


# ---------------------------------------------------------------------------
# Classification helpers
# ---------------------------------------------------------------------------

class _EndpointKind(Enum):
    """Endpoint type classification for link() sources and destinations."""
    QUEUE = 'queue'
    HANDLE = 'handle'
    PATH = 'path'
    CONNECTION = 'connection'


def _is_queue(obj) -> bool:
    """Return True if obj is a supported queue type."""
    if isinstance(obj, tuple(THREADED_QUEUES)):
        return True

    if isinstance(obj, (
        multiprocessing.queues.Queue,
        multiprocessing.queues.JoinableQueue,
        multiprocessing.queues.SimpleQueue,
    )):
        return True
    # Manager-backed proxy queues — check for queue interface
    if isinstance(obj, BaseProxy) and all(
        hasattr(obj, m) for m in ('put', 'get', 'empty')
    ):
        return True
    return False


def _classify(obj, role: DIRECTION = None) -> _EndpointKind:
    """Classify an endpoint object.

    Args:
        obj: The source or destination object to classify.
        role: ``DIRECTION.FROM`` or ``DIRECTION.TO`` — used for handle IO
            contract checks. If ``None``, contract checks are skipped.

    Returns:
        ``_EndpointKind`` enum value (QUEUE, HANDLE, PATH, or CONNECTION).

    Raises:
        TypeError: For unsupported types, QueueLink/``_LinkResult`` instances,
            or handles that do not satisfy the required IO contract.
    """
    # Reject QueueLink and _LinkResult instances immediately
    if isinstance(obj, QueueLink):
        raise TypeError(
            f"QueueLink instances are not accepted as link() endpoints. "
            f"Got: {type(obj).__name__}"
        )

    if isinstance(obj, _LinkResult):
        raise TypeError(
            f"_LinkResult instances are not accepted as link() endpoints. "
            f"Got: {type(obj).__name__}"
        )

    # Queue types (precedence: before path, since str/PathLike are also checked)
    if _is_queue(obj):
        return _EndpointKind.QUEUE

    # Connection (before generic handle check)
    if isinstance(obj, multiprocessing.connection.Connection):
        if role == DIRECTION.TO:
            raise TypeError(
                "multiprocessing.connection.Connection is not supported as a "
                "destination in this version. "
                "Pass a queue or file path/handle as the destination."
            )

        return _EndpointKind.CONNECTION

    # File paths
    if isinstance(obj, (str, PathLike)):
        return _EndpointKind.PATH

    # IO handles / duck-typed handles
    if isinstance(obj, IOBase) or (hasattr(obj, 'read') or hasattr(obj, 'write')):
        if role == DIRECTION.FROM and not hasattr(obj, 'readline'):
            raise TypeError(
                f"Source handle {type(obj).__name__!r} must support readline(). "
                "Wrap it with codecs.getreader() or provide a readline-capable handle."
            )

        if role == DIRECTION.TO and not hasattr(obj, 'write'):
            raise TypeError(
                f"Destination handle {type(obj).__name__!r} must support write()."
            )

        return _EndpointKind.HANDLE

    raise TypeError(
        f"Unsupported endpoint type: {type(obj).__name__!r}. "
        "Supported types: queue (queue.Queue, multiprocessing.Queue, etc.), "
        "file handle (IOBase or readline/write duck type), "
        "file path (str or PathLike), "
        "or multiprocessing.connection.Connection (source only)."
    )


def _normalize_destination(destination):
    """Return destination as a list, validating along the way.

    Raises:
        TypeError: For non-list iterables or QueueLink instances.
        ValueError: For empty lists or duplicate objects.
    """
    dests = []  # Will hold destination(s) as a list
    if isinstance(destination, list):
        dests = destination
    elif isinstance(destination, (tuple, set)):
        raise TypeError(
            "destination must be a single endpoint or a list of endpoints. "
            f"Got {type(destination).__name__!r}. Use a list for multiple destinations."
        )
    else:
        dests = [destination]

    if len(dests) == 0:
        raise ValueError("destination list must not be empty.")

    # Check for duplicates by identity
    seen_ids = set()
    for d in dests:
        obj_id = id(d)
        if obj_id in seen_ids:
            raise ValueError(
                f"Duplicate destination object detected: {type(d).__name__!r}. "
                "Each destination must be a distinct object."
            )
        seen_ids.add(obj_id)

    return dests


# ---------------------------------------------------------------------------
# Internal result container
# ---------------------------------------------------------------------------

class _LinkResult:  # pylint: disable=too-few-public-methods
    """Container for the components created by ``link()``.

    Not exported from the package. Users obtain an instance by calling
    ``link()`` and interact with it via ``stop()``, ``close()``, and
    ``is_alive()``.

    For queue→queue links, ``result.queue_link`` provides direct access to the
    underlying ``QueueLink`` for advanced operations (``is_empty``,
    ``is_drained``, ``get_metrics``, etc.).
    """

    def __init__(self):
        self.queue_link = None                 # QueueLink; None for adapter-only paths
        self.reader = None                     # QueueHandleAdapterReader or None
        self.writers = []                      # List of QueueHandleAdapterWriter
        self._internal_queues = []             # Queues created internally by link()
        self._drain_timeout = 5.0             # Max seconds to wait for internal queue drain

    def stop(self):
        """Stop all managed components in the correct order.

        Order:
        1. Stop the reader first (halts upstream production).
        2. Wait (bounded) for internal queues to drain into downstream stages.
        3. Stop the QueueLink publishers (if present).
        4. Stop writer adapters last (downstream sinks).
        """
        # 1. Stop reader
        if self.reader is not None:
            self.reader.close()

        # 2. Wait for internal queues to drain (best-effort, bounded)
        if self._internal_queues:
            deadline = time.monotonic() + self._drain_timeout
            for q in self._internal_queues:
                while time.monotonic() < deadline:
                    try:
                        if q.empty():
                            break
                    except OSError:
                        # Queue closed/broken (handle closed, manager shutdown, etc.)
                        break
                    time.sleep(0.01)

        # 3. Stop QueueLink
        if self.queue_link is not None:
            self.queue_link.stop()

        # 4. Stop writers
        for writer in self.writers:
            writer.close()

    def close(self):
        """Alias for ``stop()`` — consistent with adapter usage patterns."""
        self.stop()

    def is_alive(self) -> bool:
        """Return True if any managed worker thread/process is still alive."""
        if self.reader is not None and self.reader.is_alive():
            return True

        if self.queue_link is not None:
            try:
                if self.queue_link.is_alive():
                    return True
            except ProcessNotStarted:
                # QueueLink never started; treat as not alive
                pass

        for writer in self.writers:
            if writer.is_alive():
                return True

        return False


# ---------------------------------------------------------------------------
# Factory function
# ---------------------------------------------------------------------------

def link(  # pylint: disable=too-many-locals,too-many-branches,too-many-statements,protected-access
        source,
        destination,
        *,
        name: str = None,
        start_method: str = None,
        thread_only: bool = False,
        trusted: bool = False,
        wrap_when: WRAP_WHEN = WRAP_WHEN.NEVER,
        wrap_threshold: int = None,
        link_timeout: float = 0.01):
    """Wire a source and destination together automatically.

    Inspects the types of ``source`` and ``destination`` and creates the
    appropriate combination of ``QueueLink``, ``QueueHandleAdapterReader``,
    and/or ``QueueHandleAdapterWriter``.

    Args:
        source: A queue, open file/pipe handle, file path (str/PathLike), or
            ``multiprocessing.connection.Connection`` to read from.
        destination: A queue, open file/pipe handle, or file path to write to.
            May also be a ``list`` of such objects for fan-out. Tuples and
            sets are not accepted (use a list).
        name: Optional name passed to created components for logging.
        start_method: Multiprocessing start method (``'fork'``,
            ``'forkserver'``, ``'spawn'``). Defaults to system preference.
        thread_only: Force threading instead of separate processes.
        trusted: For ``Connection`` sources — if ``True``, use
            ``.recv()``/``.send()``; if ``False``, use ``.recv_bytes()``.
        wrap_when: When to wrap large messages in ``ContentWrapper`` for disk
            buffering. Only applies when a reader adapter is created.
        wrap_threshold: Byte size limit before wrapping. Only relevant when
            ``wrap_when`` is ``WRAP_WHEN.AUTO``.
        link_timeout: ``queue.get()`` timeout for ``QueueLink`` publishers.

    Returns:
        A ``_LinkResult`` instance with ``stop()``, ``close()``, and
        ``is_alive()`` interface. For queue→queue links, ``result.queue_link``
        exposes the underlying ``QueueLink``.

    Raises:
        TypeError: If either endpoint is an unsupported type, a
            ``QueueLink``/``_LinkResult`` instance, a ``Connection`` as
            destination, or if ``destination`` is a non-list iterable.
        ValueError: If ``destination`` is an empty list or contains
            duplicate objects.
    """
    # Normalise destination to a list
    dests = _normalize_destination(destination)

    # Classify source
    src_kind = _classify(source, role=DIRECTION.FROM)

    # Classify all destinations
    dest_kinds = [_classify(d, role=DIRECTION.TO) for d in dests]

    # Guard: Connection source to any handle/path destination is deferred
    if (src_kind == _EndpointKind.CONNECTION and
            any(k in (_EndpointKind.HANDLE, _EndpointKind.PATH) for k in dest_kinds)):
        raise TypeError(
            "Connection source to file handle/path destinations is not supported "
            "in this version. Use a queue as the destination when sourcing from "
            "a multiprocessing.connection.Connection."
        )

    result = _LinkResult()

    # Context for creating internal queues
    ctx = multiprocessing.get_context(start_method)

    # ------------------------------------------------------------------
    # Dispatch
    # ------------------------------------------------------------------

    all_queue_dests = all(k == _EndpointKind.QUEUE for k in dest_kinds)
    single_dest = len(dests) == 1

    if src_kind == _EndpointKind.QUEUE:
        if all_queue_dests:
            # queue → queue (or list of queues): delegate entirely to QueueLink
            dest_arg = dests[0] if single_dest else dests
            ql = QueueLink(
                source=source,
                destination=dest_arg,
                name=name,
                start_method=start_method,
                thread_only=thread_only,
                link_timeout=link_timeout,
            )
            result.queue_link = ql

        else:
            # queue → handle/path (single or mixed fan-out)
            # For each destination:
            #   - queue dest: register directly on QueueLink
            #   - handle/path dest: create internal queue + writer
            ql = QueueLink(
                source=source,
                name=name,
                start_method=start_method,
                thread_only=thread_only,
                link_timeout=link_timeout,
            )
            result.queue_link = ql

            for dest, kind in zip(dests, dest_kinds):
                if kind == _EndpointKind.QUEUE:
                    ql.write(q=dest)
                else:
                    # handle or path: create internal queue → writer
                    internal_q = ctx.Queue()
                    result._internal_queues.append(internal_q)
                    ql.write(q=internal_q)
                    writer = QueueHandleAdapterWriter(
                        queue=internal_q,
                        handle=dest,
                        name=name,
                        start_method=start_method,
                        thread_only=thread_only,
                    )
                    result.writers.append(writer)

    elif src_kind in (_EndpointKind.HANDLE, _EndpointKind.PATH, _EndpointKind.CONNECTION):

        if all_queue_dests and single_dest:
            # handle/path/connection → single queue: just a reader adapter
            reader = QueueHandleAdapterReader(
                queue=dests[0],
                handle=source,
                name=name,
                start_method=start_method,
                thread_only=thread_only,
                trusted=trusted,
                wrap_when=wrap_when,
                wrap_threshold=wrap_threshold,
            )
            result.reader = reader

        elif all_queue_dests and not single_dest:
            # handle/path/connection → [queue, queue, ...]
            # reader → internal queue → QueueLink fan-out
            # QueueLink must be fully wired before the reader starts, so that
            # data arriving in internal_q is only consumed once all destination
            # publishers are running (not by an intermediate single-dest publisher).
            internal_q = ctx.Queue()
            result._internal_queues.append(internal_q)

            ql = QueueLink(
                source=internal_q,
                destination=dests,
                name=name,
                start_method=start_method,
                thread_only=thread_only,
                link_timeout=link_timeout,
            )
            result.queue_link = ql

            reader = QueueHandleAdapterReader(
                queue=internal_q,
                handle=source,
                name=name,
                start_method=start_method,
                thread_only=thread_only,
                trusted=trusted,
                wrap_when=wrap_when,
                wrap_threshold=wrap_threshold,
            )
            result.reader = reader

        elif src_kind == _EndpointKind.CONNECTION:
            # Connection → mixed destinations: not supported in this version
            raise TypeError(
                "Connection source to mixed (queue + handle/path) destinations "
                "is not supported in this version."
            )

        elif single_dest:
            # handle/path → single handle/path: reader → internal queue → writer
            internal_q = ctx.Queue()
            result._internal_queues.append(internal_q)

            reader = QueueHandleAdapterReader(
                queue=internal_q,
                handle=source,
                name=name,
                start_method=start_method,
                thread_only=thread_only,
                trusted=trusted,
                wrap_when=wrap_when,
                wrap_threshold=wrap_threshold,
            )
            result.reader = reader

            writer = QueueHandleAdapterWriter(
                queue=internal_q,
                handle=dests[0],
                name=name,
                start_method=start_method,
                thread_only=thread_only,
            )
            result.writers.append(writer)

        else:
            # handle/path → [mixed destinations]
            # reader → internal queue → QueueLink → per-dest
            # QueueLink and all per-destination writers must be fully wired
            # before the reader starts, so no data is consumed by an
            # intermediate publisher that only knows about a subset of dests.
            internal_q = ctx.Queue()
            result._internal_queues.append(internal_q)

            ql = QueueLink(
                source=internal_q,
                name=name,
                start_method=start_method,
                thread_only=thread_only,
                link_timeout=link_timeout,
            )
            result.queue_link = ql

            for dest, kind in zip(dests, dest_kinds):
                if kind == _EndpointKind.QUEUE:
                    ql.write(q=dest)
                else:
                    fan_q = ctx.Queue()
                    result._internal_queues.append(fan_q)
                    ql.write(q=fan_q)
                    writer = QueueHandleAdapterWriter(
                        queue=fan_q,
                        handle=dest,
                        name=name,
                        start_method=start_method,
                        thread_only=thread_only,
                    )
                    result.writers.append(writer)

            reader = QueueHandleAdapterReader(
                queue=internal_q,
                handle=source,
                name=name,
                start_method=start_method,
                thread_only=thread_only,
                trusted=trusted,
                wrap_when=wrap_when,
                wrap_threshold=wrap_threshold,
            )
            result.reader = reader

    return result



# -*- coding: utf-8 -*-
"""Wrapper to make async writing to a pipe more reliable across processes"""
from __future__ import unicode_literals

from kitchen.text.converters import to_bytes
from kitchenpatch import getwriter

from .exceptionhandler import ExceptionHandler


def writeout(pipe, output_prefix):
    """Easily write to your favorite pipe or handle with current content

    Args:
        pipe (pipe): A system pipe/file handle to write output to
        output_prefix (string): A string to prepend to each line

    Returns:
        function
    """
    # TODO Validate the pipe somehow

    def func(line):
        pipe_writer = getwriter("utf-8")(pipe)
        output = f'{output_prefix}{line}'

        try:
            pipe_writer.write(output)

        except TypeError:
            # Shenanigans with unicode
            try:
                pipe_writer.write(to_bytes(output))
            except TypeError:
                pipe.write(str(output))
            except Exception as exc:
                raise ExceptionHandler(exc, f'Crazy pipe writer stuff: {exc}')

        except ValueError as exc:
            raise ExceptionHandler(exc, f'writeout caught odd error: {exc}')

        finally:
            pipe_writer.flush()
            pipe.flush()

    return func



# -*- coding: utf-8 -*-
"""Custom queue-pipe adapter to read from thread-safe queues and write their
contents to a pipe"""
from __future__ import unicode_literals

import io
import os
import logging

from queue import Empty
from os import PathLike
from typing import Union, get_args

from .contentwrapper import ContentWrapper
from .queue_handle_adapter_base import MessageCounter
from .queue_handle_adapter_base import _QueueHandleAdapterBase
from .timer import Timer
from .common import (
    safe_get,
    UNION_SUPPORTED_EVENTS,
    UNION_SUPPORTED_LOCKS,
    UNION_SUPPORTED_QUEUES,
    DIRECTION,
    UNION_SUPPORTED_IO_TYPES,
    UNION_SUPPORTED_PATH_TYPES)


# Private class only intended to be used by ProcessRunner
# Works around (https://bryceboe.com/2011/01/28/
# the-python-multiprocessing-queue-and-large-objects/ with large objects)
# by using ContentWrapper to buffer large lines to disk
class QueueHandleAdapterWriter(_QueueHandleAdapterBase):
    """Custom manager to read messages from a queue and write them to a file or pipe
    """
    def __init__(self,
                 queue: UNION_SUPPORTED_QUEUES,
                 *,  # End of positional arguments
                 handle: UNION_SUPPORTED_IO_TYPES=None,
                 name: str=None,
                 log_name: str=None,
                 start_method: str=None,
                 thread_only: bool=None,
                 trusted: bool=False):
        """Custom manager to read messages from a queue and write them to a file or pipe

        Args:
            queue: Queue to retrieve messages from
            handle: File name, handle, or pipe to write messages to
            name: Optional name for this reader
            log_name: Optional name for this reader in log lines
            start_method: Explicit multiprocessing start method to use
            thread_only: Force the adapter to use a thread rather than process
            trusted: Whether to trust Connection objects; True uses .send/.recv, False
                send_bytes/recv_bytes when reading from multiprocessing.connection.Connections
        """
        # Initialize the parent class
        super().__init__(queue=queue,
                         subclass_name=__name__,
                         queue_direction=DIRECTION.TO,
                         name=name,
                         handle=handle,
                         log_name=log_name,
                         start_method=start_method,
                         thread_only=thread_only,
                         trusted=trusted)

    @staticmethod
    def queue_handle_adapter(*,  # All named parameters are required keyword arguments
                             name: str,
                             handle: UNION_SUPPORTED_IO_TYPES,
                             queue: UNION_SUPPORTED_QUEUES,
                             queue_lock: UNION_SUPPORTED_LOCKS,
                             stop_event: UNION_SUPPORTED_EVENTS,
                             messages_processed: MessageCounter,
                             trusted: bool,
                             **kwargs):
        """Copy lines from a local multiprocessing.JoinableQueue into a pipe

        Runs in a separate process, started by __init__. Does not close an open
        pipe or handle when done writing.

        Args:
            name: Name to use in logging
            handle: Handle/pipe/path to write to
            queue: Queue to write to
            queue_lock: Lock used to indicate a write is in progress
            stop_event: Used to determine whether to stop the process
            messages_processed: Number of elements moved from the queue to handle
            trusted: Whether to trust Connection objects
        """
        def open_location(location: Union[str, PathLike], line) -> \
                [io.TextIOWrapper, io.BufferedWriter]:
            """Open a location string/Path and return a normal IO handle."""
            if hasattr(line, 'decode'):
                # Open the output file in binary mode
                return open(location, mode='w+b')

            # Otherwise open as a text file
            return open(location, mode='w+')  # pylint: disable=unspecified-encoding

        def flush(file_handle):
            """Simple function to push content to disk"""
            if hasattr(file_handle, 'flush'):
                file_handle.flush()

            if hasattr(file_handle, 'fileno'):
                os.fsync(file_handle.fileno())

        log = logging.getLogger(f'{__name__}.queue_handle_adapter.{name}')
        log.addHandler(logging.NullHandler())

        log.info('Starting writer process')
        if hasattr(handle, 'closed') and handle.closed:
            log.warning('Handle is already closed')

        else:
            flush_timer = Timer(interval=1)  # Flush to disk at least once a second

            # Make comparisons easier/faster when checking for an open file
            # get_args syntax used for Python 3.8-3.12 compatibility
            #   https://stackoverflow.com/a/64643971
            handle_ready = True
            if isinstance(handle, get_args(UNION_SUPPORTED_PATH_TYPES)):
                handle_name = handle
                handle_ready = False

            # Handle type
            is_handle_bin = None

            # Loop over available lines until asked to stop
            while True:
                try:
                    line = safe_get(queue, timeout=0.05, stop_event=stop_event)

                    # Extract the content if the line is in a ContentWrapper
                    if line is not None:
                        content = line.value if isinstance(line, ContentWrapper) else line
                        is_content_bin = hasattr(content, 'decode')

                        # Lazily open the file handle if it is not already open
                        if not handle_ready:
                            handle = open_location(handle_name, line)  # pylint: disable=possibly-used-before-assignment
                            handle_ready = True

                        # Determine if the handle is binary
                        is_handle_bin = 'b' in handle.mode

                        # Write content into the file
                        log.info('Writing line to %s', name)
                        if is_handle_bin:
                            handle.write(content if is_content_bin else content.encode('utf-8'))
                        else:
                            handle.write(content.decode('utf-8') if is_content_bin else content)

                    # Indicate we finished processing a record
                    messages_processed.increment()

                    # Signal to the queue that we are done processing the line
                    if hasattr(queue, 'task_done'):
                        queue.task_done()

                    # Flush the pipe to make sure it gets to the process
                    if flush_timer.interval():
                        flush(handle)

                    # Exit if we are asked to stop
                    if stop_event.is_set():
                        log.info("Writer asked to stop")
                        break

                except Empty:
                    log.debug("No line currently available for %s", name)

                    # Exit if we are asked to stop
                    if stop_event.is_set():
                        log.info("Writer asked to stop")
                        break

        # Do a final flush
        flush(handle)

        # Close the handle if we opened it
        if 'handle_name' in locals() and hasattr(handle, 'close'):
            handle.close()

        # A bit of info
        log.info('Writer wrote %d messages', messages_processed.value)

        # Clean up references
        # Prevent "UserWarning: ResourceTracker called reentrantly for resource cleanup,
        # which is unsupported. The semaphore object '/<name>' might leak."
        queue_lock = None
        stop_event = None
        messages_processed = None

        log.info("Writer sub-process complete")



# -*- coding: utf-8 -*-
"""QueueLink allows you to easily connect queues together"""
# Define version variable
from importlib_metadata import version, packages_distributions, PackageNotFoundError

from queuelink import common

from queuelink.common import DIRECTION
from queuelink.queuelink import QueueLink
from queuelink.timer import Timer
from queuelink.common import UNION_SUPPORTED_QUEUES
from queuelink.common import safe_get

# Pipe adapters
from queuelink.contentwrapper import ContentWrapper
from queuelink.contentwrapper import WRAP_WHEN
from queuelink.writeout import writeout
from queuelink.queue_handle_adapter_reader import QueueHandleAdapterReader
from queuelink.queue_handle_adapter_writer import QueueHandleAdapterWriter

# Factory function
from queuelink.link import link

try:
    packages = packages_distributions()
    package_name = packages[__name__][0]
    __version__ = version(package_name)
except (PackageNotFoundError, KeyError):
    # package is not installed
    pass



# -*- coding: utf-8 -*-
"""Shared constants and functions"""
import queue
import random
import time

from enum import Enum
from os import PathLike
from queue import Empty
from threading import Event as t_Event
from threading import Lock as t_Lock
from typing import IO, List, Union

# Multiprocessing imports
import multiprocessing
from multiprocessing import Event as mp_Event
from multiprocessing import Lock as mp_Lock
from multiprocessing import queues as mp_queue_classes
from multiprocessing.managers import BaseProxy

# Internal imports; keep to minimum to prevent cyclical dependencies
from .timer import Timer

class DIRECTION(Enum):
    """Directions enum (from source/to destination)"""
    def __str__(self):
        return str(self.value)

    FROM = 'source'
    TO = 'destination'

# Supported events
UNION_SUPPORTED_EVENTS = Union[
    t_Event,
    mp_Event,
    BaseProxy
]

# Supported locks
UNION_SUPPORTED_LOCKS = Union[
    t_Lock,
    mp_Lock,
    BaseProxy
]

# Ways to start a process
PROC_START_METHODS = ['fork', 'forkserver', 'spawn']

# Module, Class, Max size
QUEUE_TYPE_LIST = [
    ('manager', 'Queue', None),
    ('manager', 'JoinableQueue', None),
    ('multiprocessing', 'Queue', None),
    ('multiprocessing', 'JoinableQueue', None),
    ('multiprocessing', 'SimpleQueue', None),
    ('queue', 'Queue', None),
    ('queue', 'LifoQueue', None),
    ('queue', 'PriorityQueue', None),
    ('queue', 'SimpleQueue', None)
]

# List of queue types that are threaded (vs multiprocessing)
THREADED_QUEUES = [
    queue.Queue,
    queue.LifoQueue,
    queue.PriorityQueue,
    queue.SimpleQueue
]

# List of PriorityQueue types
PRIORITY_QUEUES = [
    queue.PriorityQueue
]

# List of SimpleQueue types
SIMPLE_QUEUES = [
    queue.SimpleQueue,
    mp_queue_classes.SimpleQueue
]

UNION_MULTIPROCESSING_QUEUES = Union[
    multiprocessing.queues.Queue,
    multiprocessing.queues.JoinableQueue,
    multiprocessing.queues.SimpleQueue
]

# Union type for typing support
# pylint: disable=invalid-name
UNION_SUPPORTED_QUEUES = Union[
    queue.Queue,
    queue.LifoQueue,
    queue.PriorityQueue,
    queue.SimpleQueue,
    multiprocessing.queues.Queue,
    multiprocessing.queues.JoinableQueue,
    multiprocessing.queues.SimpleQueue,
    BaseProxy
]

# Union type for IO (reader/writer)
# pylint: disable=invalid-name
UNION_SUPPORTED_IO_TYPES = Union[
    IO,
    str,
    PathLike
]

# Union type for file pointer types (reader/writer)
# pylint: disable=invalid-name
UNION_SUPPORTED_PATH_TYPES = Union[
    str,
    PathLike
]


def is_threaded(queue_list: Union[List, UNION_SUPPORTED_QUEUES]) -> bool:
    """Check whether one or more queues are threaded

    Args:
         queue_list: Any queue or list of queues (any type)

    Returns:
        True if any queue in the given list is threaded.
    """
    if not isinstance(queue_list, list):
        queue_list = [queue_list]

    for q_inst in queue_list:
        if isinstance(q_inst, tuple(THREADED_QUEUES)):
            return True

    return False

def new_id() -> str:
    """Create a reasonably unique ID string.

    Returns:
        A 6-character unique ID.
    """
    return ''.join([random.choice(  # nosec
               '0123456789ABCDEF') for x in range(6)])

def safe_get(queue_obj: UNION_SUPPORTED_QUEUES,
             block: bool = True,
             timeout: float = 0,
             stop_event: Union[t_Event, mp_Event] = None,
             cycle_time: float = 0.005) -> any:
    """Queue ``get`` implementation that implements partial timeout for all Queue types including
    SimpleQueues.

    CAUTION: Can deadlock if a SimpleQueue is read by multiple consumers.

    Args:
        queue_obj: Any queue object
        block: Block until queue has an element to return (default True)
        timeout: Timeout in seconds; if block is True, raise Empty after this duration
        stop_event: Event to force stop
        cycle_time: Cycle time in seconds of blocking logic for SimpleQueues

    Returns:
        Returns a queue element or None

    Raises:
        queue.Empty
    """
    # Handle cycle time with the stop_event if present
    def wait(time_out: float=0):
        """Return False when timeout expires or True if the stop_event is set

        Args:
            time_out: Timeout in seconds

        Returns:
            False when timeout expires or True if the stop_event is set
        """
        if not stop_event:
            time.sleep(time_out)
            return False

        return stop_event.wait(time_out)

    try:
        return queue_obj.get(timeout=timeout)

    # get(block=, timeout=) method signature doesn't exist for SimpleQueues
    # Alternative implementation to simulate the behavior
    except TypeError:
        timer = Timer(interval=timeout)

        while True:
            # Asked to stop, so just stop
            if stop_event and stop_event.is_set():
                raise Empty

            # Alternate timer
            if timer.interval():
                raise Empty

            # Check if the queue has a get_nowait method, equivalent to get(block=False)
            # This probably won't be available if get(timeout=) failed, but checking anyway
            if hasattr(queue_obj, "get_nowait"):
                try:
                    return queue_obj.get_nowait()

                except Empty:
                    if not block:
                        raise Empty

                    if wait(time_out=cycle_time):  # True if stop event was set
                        raise Empty

                    continue

                except (EOFError, BrokenPipeError):
                    raise Empty

            # SimpleQueues don't have a get_nowait method/mechanism
            # Try not to get stuck, but can't guarantee that another thread hasn't
            # grabbed one
            if queue_obj.empty():
                if not block:
                    raise Empty

                if wait(cycle_time):  # True if stop event was set
                    # Raise empty if the stop_event is set to break out of the loop
                    raise Empty

                # Timeout happened, begin the next while cycle
                continue

            # SimpleQueue with an item in the queue
            # This can deadlock if there's another reader of this same queue
            try:
                return queue_obj.get()
            except (EOFError, BrokenPipeError):
                raise Empty

    except (EOFError, BrokenPipeError):
        raise Empty

    finally:
        queue_obj = None
        stop_event = None



# -*- coding: utf-8 -*-
"""
Exception management classes
"""
from __future__ import unicode_literals

import logging
import sys
import traceback


# pylint: disable=super-init-not-called
# User-defined exceptions don't call their super initializer


class SIGINTException(Exception):
    """Represents a ctrl-c interrupt"""
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class ProcessNotStarted(Exception):
    """Raise if a publishing QueueLink hasn't been started, but a method has been
    called that depends on the target process running.
    """
    def __init__(self, value=""):
        """Arguments must be option to prevent triggering
        https://bugs.python.org/issue15440 when raised in _Command"""
        self.errno = 4
        self.value = value

    def __str__(self):
        return repr(self.value)


class HandleAlreadySet(Exception):
    """Raise if a pipe adapter has already been configured with a pipe handle"""
    def __init__(self, value=""):
        """Arguments must be option to prevent triggering
        https://bugs.python.org/issue15440 when raised in _Command"""
        self.errno = 5
        self.value = value

    def __str__(self):
        return repr(self.value)


class HandleNotSet(Exception):
    """Raise if a pipe adapter has not been configured with a pipe handle, but
    a call requires one have been set
    """
    def __init__(self, value=""):
        """Arguments must be option to prevent triggering
        https://bugs.python.org/issue15440 when raised in _Command"""
        self.errno = 6
        self.value = value

    def __str__(self):
        return repr(self.value)


class ExceptionHandler(Exception):
    """Exception management

    TODO: Add additional detail
    """
    def __repr__(self):
        return self.errmsg

    def __str__(self):
        return self.errmsg

    def __init__(self, error, message=None):
        self.exc_type, self.exc_obj, self.exc_tb = sys.exc_info()
        # fname = os.path.split(self.exc_tb.tb_frame.f_code.co_filename)[1]
        self.err_type = type(error).__name__
        self.error_text = str(error)

        log = logging.getLogger(__name__)
        log.addHandler(logging.NullHandler())

        if message is not None:
            log.error(message)

        template = "An exception of type {0} occurred. Error message:\n{1}"
        self.errmsg = template.format(self.err_type, self.error_text)
        self.errmsg += "\n"
        log.error(self.errmsg)

        errargmsg = f"{self.err_type} {type(error).__name__} arguments:\n{error.args:2!r}"
        errargmsg += "\n"
        log.error(errargmsg)

        tbmsg = self.err_type+" traceback (most recent call last):"
        log.error(tbmsg)
        log.error("".join(traceback.format_tb(self.exc_tb)))



# -*- coding: utf-8 -*-
"""
Representation of content for a queue where the values may exceed the native
pipe size.
"""
import math
import os
import tempfile

from codecs import getreader
from enum import Enum, auto
from io import IOBase as file
from typing import Union

from kitchen.text.converters import to_bytes
from kitchenpatch import getwriter

from .classtemplate import ClassTemplate
from .timer import Timer


def get_len(obj: any) -> int:
    """Calculate the byte length of a given object.

    Args:
        obj: The object to calculate the byte length for.

    Returns:
        The byte length of the given object.
    """
    if isinstance(obj, str):
        subject_byte_count = len(obj.encode('utf8'))
    else:
        subject_byte_count = len(to_bytes(obj))

    return subject_byte_count


def of_bytes_length(subject: str, length: int, round_func_name='floor') -> str:
    """Generate an object of the given length out of the subject.

    Calculates the length of "subject" in bytes, then generates a longer string as multiples
    of ``subject`` that comes in at or under ``length``. See next note for "at or above".

    Set ``round_func_name`` to ``ceil`` to generate a longer string that comes in at or above
    ``length``.

    Args:
        subject: The baseline object to multiply to create the output object
        length: The length to generate.
        round_func_name: The name of the function used to round the length.

    Returns:
        An extended string based on the ``subject``.
    """
    subject_byte_count = get_len(subject)

    # If the length is less than the byte count of the subject (say length 2 but the byte
    # count is 4) return the subject unmodified
    if length < subject_byte_count:
        return subject

    # Rounding function
    round_func = getattr(math, round_func_name)

    # subject_byte_count = 4
    # length = 22
    # math.floor(int(length=22) / subject_byte_count=4) = math.floor(5.5) = 5
    # joined length = 20
    subject_array = [subject] * int(round_func(int(length) / subject_byte_count))
    return "".join(subject_array)


class TYPES(Enum):
    """Enum indicating whether a wrapped item is in memory or a file"""
    DIRECT = 0
    FILE = 1


class WRAP_WHEN(Enum):  # pylint: disable=invalid-name
    """Enum for deciding when to use ContentWrapper
    Not used directly inside ContentWrapper

    AUTO: Use a ContentWrapper only when content is over ContentWrapper.THRESHOLD (or local size)
    ALWAYS: All content is wrapped
    NEVER: No content is wrapped
    """
    AUTO = auto()
    ALWAYS = auto()
    NEVER = auto()


class ContentWrapper(ClassTemplate):
    """
    Representation of content for a queue where the values may exceed the
    native pipe size.

    Store data with
        cw = ContentWrapper("My text data")
        cw.value = "Updated text data"

    Access data with
        print("My data: {cw.value}")
        print("My data: {cw}")
    """

    # Need to figure out a way to do this automatically
    THRESHOLD = 2**14  # 2**14 = 16,384

    def __getattr__(self, attr):
        """When necessary pull the 'value' from a buffer file"""
        log = object.__getattribute__(self, "_log")

        # pylint: disable=no-else-return
        # The 'else' can be hit, so it is not superfluous
        if attr == "value" \
                and object.__getattribute__(self, "storage_type") == TYPES.FILE:
            log.debug("Pulling value from buffer file")
            return self._get_value_from_file()
        else:
            log.debug("Pulling value from memory")
            return object.__getattribute__(self, attr)

    def __setattr__(self, attr, val):
        """When necessary, save the 'value' to a buffer file"""
        if attr == "value":
            # Within the threshold size limit or not forced to use a file
            if get_len(val) < self.threshold and not self._is_explicit_file():
                self._log.debug("Storing value to memory")
                object.__setattr__(self, attr, val)

                # Delete temp file if we had previously used a file
                self._delete_temp_file()

            # Larger than what a queue value can hold
            # due to pipe limits, store value in a temp file
            else:
                # Clear existing in-memory value
                if hasattr(object, 'value'):
                    del self.value

                # Set the type
                object.__setattr__(self, "storage_type", TYPES.FILE)

                # pylint: disable=consider-using-with
                # We explicitly do not want to close the tempfile automatically
                if not self.location_handle:
                    object.__setattr__(self, "location_handle",
                                       tempfile.NamedTemporaryFile(delete=False))
                handle = object.__getattribute__(self, "location_handle")
                object.__setattr__(self, "location_name", handle.name)
                writer = getwriter("utf-8")(handle)

                self._log.info("Writing value into buffer file %s",
                               handle.name)
                stopwatch = Timer()
                writer.write(val)
                writer.flush()
                os.fsync(writer.fileno())
                lap = stopwatch.lap()
                self._log.info("Finished writing value into buffer file in "
                               "%.1f seconds", lap)

        # Not assigning to self.value
        else:
            object.__setattr__(self, attr, val)

    def __getstate__(self):
        """Being Pickled"""
        self._log.debug("Being pickled")

        # Close the buffer file if needed
        if isinstance(self.location_handle, file):
            self.location_handle.close()

        state = self.__dict__.copy()
        del state['location_handle']
        del state['_log']  # Delete the logger instance

        # Prevent __del__ from deleting the buffer file
        # Needs to come after we've created the state copy so
        # this doesn't persist after un-pickling
        self.being_serialized = True

        return state

    def __setstate__(self, state):
        """Being un-Pickled, need to restore state"""

        self.__dict__.update(state)
        self.location_handle = None

        # Reestablish the logger
        self._initialize_logging(__name__)
        self._log.debug("Being un-pickled")

        if self.location_name is not None:
            # pylint: disable=consider-using-with
            # We explicitly do not want to close the file automatically
            self.location_handle = open(self.location_name, "r+b")

    def __del__(self):
        """When used, close any open file handles on object destruction"""
        self._log.debug("Object being deleted")
        self._delete_temp_file()

    def __len__(self):
        return len(self.value)

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.value}')"

    def __str__(self):
        if isinstance(self.value, str):
            return self.value

        if isinstance(self.value, bytes):
            return self.value.decode('utf8')

        return str(self.value)

    def __bytes__(self):
        if isinstance(self.value, str):
            return self.value.encode('utf8')

        if isinstance(self.value, bytes):
            return self.value

        return bytes(self.value)

    def __eq__(self, other):
        return self.value == other

    def __init__(self, val, threshold: int = None, storage_type: TYPES = None):
        self._log = None
        self._initialize_logging(__name__)

        self.threshold = self.THRESHOLD if threshold is None else threshold

        # Set the type
        if storage_type:
            if isinstance(storage_type, TYPES):
                self.storage_type = storage_type
                self.type_explicit = True
            else:
                raise TypeError(f'Given type "{storage_type}" is not from TYPES')
        else:
            self.storage_type = TYPES.DIRECT
            self.type_explicit = False

        # Used only if this is stored in a file
        self.location_handle: file = None
        self.location_name: str = None
        self.being_serialized: bool = False

        # Store the initial value
        self.value = val

    def _create_buffer(self):
        pass

    def _is_explicit_file(self):
        if self.type_explicit and self.storage_type == TYPES.FILE:
            return True

        return False

    def _get_value_from_file(self):
        handle = object.__getattribute__(self, "location_handle")
        reader = getreader("utf-8")(handle)
        handle.seek(0)

        stopwatch = Timer()
        content = reader.read()
        lap = stopwatch.lap()
        self._log.info("Finished reading value into buffer file in %.1f "
                       "seconds", lap)

        return content

    def _delete_temp_file(self):
        """Delete the temp file if set"""
        if isinstance(self.location_handle, file):
            self._log.debug('Closing temp file')
            self.location_handle.close()

        # Delete any files on disk
        if self.location_name and not self.being_serialized:
            self._log.debug('Deleting temp file')
            os.remove(self.location_name)
            self.location_handle = None
            self.location_name = None


def conditional_wrap(content: any,
                     wrap_when: WRAP_WHEN=WRAP_WHEN.AUTO,
                     wrap_threshold: int=ContentWrapper.THRESHOLD) -> Union[any, ContentWrapper]:
    """Method to wrap content only if it meets the given conditions.

    Typically to keep values that would be stored in memory from being wrapped.

    Passes wrap_threshold as the file threshold into the ContentWrapper instance.

    Args:
        content: Object to be conditionally wrapped.
        wrap_when: Condition when the wrapping should occur.
        wrap_threshold: Size threshold over which the wrapping should occur.

    Returns:
        The wrapped content.
    """

    # Never wrap
    if wrap_when == WRAP_WHEN.NEVER:
        return content

    # Always wrap
    if wrap_when == WRAP_WHEN.ALWAYS:
        return ContentWrapper(content, threshold=wrap_threshold)

    # Wrap only over certain thresholds
    if wrap_when == WRAP_WHEN.AUTO \
            and get_len(content) > wrap_threshold:
        return ContentWrapper(content, threshold=wrap_threshold)

    return content



# -*- coding: utf-8 -*-
"""Class to house time-related convenience functions."""
import datetime
import math  # pylint: disable=no-name-in-module
import time


class Timer:
    """Used to help time events."""
    SEC_TO_MICRO = 10**6
    SEC_TO_NANO = 10**9

    def __init__(self, interval: float=10):
        """Timing events: Establish start time reference for lap and interval

        Args:
            interval: Reference duration for ``interval`` method. Supports resolution to
                microseconds. Default is 10 seconds.
        """
        self.interval_period = interval
        self.start_time = self.now()
        self.start_moment_ns = time.perf_counter_ns()
        self.last_interval_count = 0

    @staticmethod
    def now() -> float:
        """Returns the current Unix epoch time

        Returns:
            Current Unix epoch time in integer seconds with decimal resolution to microseconds.
        """
        dt = datetime.datetime.now()
        current = time.mktime(dt.timetuple()) + dt.microsecond / float(Timer.SEC_TO_MICRO)

        return float(current)

    @staticmethod
    def now_micro() -> int:
        """Returns the current Unix epoch time in microseconds as an integer

        Returns:
            Current Unix epoch time in microseconds.
        """
        current = time.mktime(datetime.datetime.now().timetuple()) * Timer.SEC_TO_MICRO \
                  + datetime.datetime.now().microsecond

        return int(current)

    def lap_ns(self) -> int:
        """Return nanoseconds since this instance was created.

        Returns:
            Nanoseconds since this instance was created
        """
        return time.perf_counter_ns() - self.start_moment_ns

    def lap(self) -> float:
        """Return seconds since this instance was created.

        Includes decimal to nanosecond resolution, however, the division of nanoseconds is
        unreliable. Use ``lap_ns`` if high precision is required.

        Returns:
            Seconds since this instance was created (to nanoseconds)
        """
        return self.lap_ns() / Timer.SEC_TO_NANO

    def interval(self) -> bool:
        """Return True if we have exceeded the interval since we started or
        last called ``interval``.

        Returns:
            True if we have exceeded ``interval`` since we started or last called ``interval``,
                False otherwise.
        """

        # Get the current lap time
        lap = self.lap_ns()  # Nanoseconds
        interval = int(self.interval_period * Timer.SEC_TO_NANO)

        # How many intervals have elapsed since this instance was created?
        interval_count_float = lap / interval  # integer division

        # Round down the number of intervals to a whole number
        # pylint: disable=c-extension-no-member
        interval_count_floor = int(math.floor(interval_count_float))
        # pylint: enable=c-extension-no-member

        # If an interval has passed since the last one was recorded,
        # return true
        # pylint: disable=no-else-return
        if interval_count_floor > self.last_interval_count:
            self.last_interval_count = interval_count_floor
            return True
        else:
            return False



# -*- coding: utf-8 -*-
"""Manages the pull/push with queues"""
from __future__ import unicode_literals

from typing import Dict, List, Union

import functools
import logging

from builtins import str as text
from inspect import signature
import queue
from queue import Empty, Full
from threading import Thread  # For non-multi-processing queues

# Multiprocessing imports
import multiprocessing
from ctypes import c_int

# Internal imports
from .classtemplate import ClassTemplate
from .exceptionhandler import ProcessNotStarted
from .metrics import Metrics, MetricType
from .common import DIRECTION
from .common import new_id, is_threaded, safe_get
from .common import (
    PRIORITY_QUEUES,
    SIMPLE_QUEUES,
    UNION_MULTIPROCESSING_QUEUES,
    UNION_SUPPORTED_EVENTS,
    UNION_SUPPORTED_LOCKS,
    UNION_SUPPORTED_QUEUES)


def validate_direction(func):
    """Decorator to check that 'direction' is an acceptable value.

    One of DIRECTION.FROM or DIRECTION.TO.

    :raises TypeError
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):

        # Extract the function signature so we can check for direction
        sig = signature(func)
        bound = sig.bind_partial(*args, **kwargs)

        # Validate direction if present
        if "direction" in bound.arguments:
            arg_direction = bound.arguments['direction']

            # If the direction is valid, just keep going
            if arg_direction in DIRECTION:
                pass

            # If the direction is invalid, throw an error
            else:
                raise TypeError(
                    "destination must be a value from DIRECTION")

        # Call the normal function
        return func(*args, **kwargs)

    return wrapper


class LimitedLengthQueue(object):
    """Wrapper for an existing Queue to keep the length limited.

    NOT thread/process safe, only intended to be used by a single producer (i.e., Publishers).

    NOTE: This class is not currently used in the codebase. The metrics path previously used
    it but was replaced by bounded Queue(maxsize=) + put_nowait() in FEAT-002. Retained as a
    general-purpose utility in case future callers need a bounded blocking wrapper.
    """
    def __init__(self,
                 queue_instance: multiprocessing.queues.Queue,
                 max_size: int=0,
                 start_method: str=None):
        # Which multiprocess context to use
        self.multiprocessing_ctx = multiprocessing.get_context(start_method)

        self._queue = queue_instance
        self.max_size = max_size
        self._queue_length = self.multiprocessing_ctx.Value(c_int, 0)

        if not self._queue.empty():
            raise ValueError('Cannot start with a non-empty Queue!')

    def put(self, *args, **kwargs):
        """Add an element to the queue"""
        while self._queue_length.value >= self.max_size:
            self.get()

        self._queue.put(*args, **kwargs)
        with self._queue_length.get_lock():
            self._queue_length.value += 1

    def get(self, *args, **kwargs):
        """Retrieve an element from the queue

        :throws Empty
        """
        value = self._queue.get(*args, **kwargs)
        with self._queue_length.get_lock():
            self._queue_length.value -= 1

        return value

    def qsize(self):
        """Returns the number of elements in the queue

        Based on the wrapper's counter; does not rely on the underlying queue implementation.
        """
        return self._queue_length.value

    def __getattr__(self, item):
        if item in ('_queue', '_queue_length', 'max_size', 'get', 'put'):
            return self.__getattribute__(item)

        return getattr(self._queue, item)


class QueueLink(ClassTemplate):
    """Manages publishing from source and client queues"""

    def __init__(self,
                 source: UNION_SUPPORTED_QUEUES=None,
                 destination: Union[UNION_SUPPORTED_QUEUES, List[UNION_SUPPORTED_QUEUES]]=None,
                 *,  # Remainder are keyword arguments
                 name: str=None,
                 log_name: str=None,
                 start_method: str=None,
                 thread_only: bool=False,
                 link_timeout: float=0.01):
        """Manages the pull/push with queues

        Args:
            source: Queue to act as the source
            destination: Queue or list of queues to act as a destination
            name: Appended to class name in log records (``cl.name``)
            log_name: Appended to class name.name (``cl.name.log_name``)
                or just class name if ``name`` is not provided
            start_method: For multiprocess use: fork, spawn or forkserver
            thread_only: Only use threads, not separate processes, for links
            link_timeout: Tune the queue.get(timeout=link_timeout) value; default 0.01 sec
        """
        # Unique ID
        # Not used for cryptographic purposes, so excluding from Bandit
        self.id = new_id()

        self.name = name
        self.log_name = log_name

        self._initialize_logging_with_log_name(__name__)

        # Timeout for publisher queue.get calls
        self.link_timeout = link_timeout

        # Thread only usage (no multiprocessing)
        self.thread_only = thread_only

        # Start method for process-based use
        self.start_method = start_method

        # Use the correct types when creating multithreaded/process Events
        self.multiprocess_ctx = multiprocessing.get_context(self.start_method)

        # Indicate whether we have ever been started (Event())
        self.started = self.Event()

        # List of locks allows decoupled IO while also preventing the set of
        # queues from changing during IO operations
        self.queues_lock = self.Lock()
        self.last_queue_id = 0
        self.client_queues_source = {}
        self.client_queues_destination = {}
        self.client_pair_publishers = {}
        self.publisher_stops = {}

        # Metrics — two queues to avoid feeder-thread races:
        # Thread publishers use a threading queue (synchronous put, no feeder thread).
        # Process publishers use a multiprocessing queue (cross-process safe), created
        # lazily on first process-based publisher start to avoid unnecessary OS resources.
        self._metrics_queue_thread = queue.Queue(maxsize=100)
        self._metrics_queue_process = None

        # Short-circut usage
        if source:
            self.register_queue(q=source, direction=DIRECTION.FROM)

        if destination:
            if not isinstance(destination, list):
                destination = [destination]

            for q_inst in destination:
                self.register_queue(q=q_inst, direction=DIRECTION.TO)


    # Class contains Locks and Queues which cannot be pickled
    def __getstate__(self):
        """Prevent _QueueHandleAdapterBase from being pickled across Processes

        Raises:
            Exception
        """
        raise TypeError("Don't pickle me!")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def close(self) -> None:
        """Remove managed objects started within QueueLink"""

        # Make sure processes are stopped
        if hasattr(self, 'stop') and hasattr(self, 'started'):
            if self.started:
                self.stop()

        # Unset these variables to release any objects they held
        unset_list = ['_metrics_queue_thread',  # Metrics (thread publishers)
                      '_metrics_queue_process',  # Metrics (process publishers)
                      'started',  # Started event
                      'queues_lock',  # Queues lock
                      'publisher_stops',  # Publisher stop events
                      'client_queues_source',  # Queue references
                      'client_queues_destination',  # Queue references
                      'client_pair_publishers']  # Queue references

        for var in unset_list:
            if hasattr(self, var):
                setattr(self, var, None)

    def _initialize_logging_with_log_name(self, class_name: str):
        """Need to reverse the print order of log_name and name"""
        if hasattr(self, '_log'):
            if self._log is not None:
                return

        # Make a helpful name
        if self.name is None:
            name = class_name
        else:
            name = f'{class_name}.{self.name}'

        if self.log_name is not None:
            name = f'{name}.{self.log_name}'

        self._log = logging.getLogger(name)
        self.add_logging_handler(logging.NullHandler())

    def Event(self, *args, **kwargs) -> UNION_SUPPORTED_EVENTS:  # pylint: disable=invalid-name
        """Create a context-appropriate Event"""
        return self.multiprocess_ctx.Event(*args, **kwargs)

    def Lock(self, *args, **kwargs) -> UNION_SUPPORTED_LOCKS:  # pylint: disable=invalid-name
        """Create a context-appropriate Lock"""
        return self.multiprocess_ctx.Lock(*args, **kwargs)

    def Process(self, *args, **kwargs) -> multiprocessing.Process:  # pylint: disable=invalid-name
        """Create a context-appropriate Process"""
        return self.multiprocess_ctx.Process(*args, **kwargs)

    def Queue(self, *args, **kwargs) -> UNION_MULTIPROCESSING_QUEUES:  # pylint: disable=invalid-name
        """Create a context-appropriate Queue"""
        return self.multiprocess_ctx.Queue(*args, **kwargs)

    def _get_process_metrics_queue(self) -> UNION_MULTIPROCESSING_QUEUES:
        """Return the metrics queue for process-based publishers, creating it on first call.

        Deferred creation avoids allocating OS-level pipe resources on QueueLink instances
        that only ever use thread-based publishers.
        """
        if self._metrics_queue_process is None:
            self._metrics_queue_process = self.multiprocess_ctx.Queue(maxsize=100)

        return self._metrics_queue_process

    def stop(self) -> None:
        """Use to stop somewhat gracefully"""
        self._stop_publishers()

    @staticmethod
    def _publisher(stop_event: UNION_SUPPORTED_EVENTS,
                   source_id: str,
                   source_queue: UNION_SUPPORTED_QUEUES,
                   dest_queues_dict: Dict[str, UNION_SUPPORTED_QUEUES],
                   timeout: float,
                   metrics_queue: UNION_SUPPORTED_QUEUES,
                   metric_interval: int = 100,
                   name: str = None) -> None:
        """Move messages from the source queue to the destination queue

        Args:
            stop_event: Used to determine whether to stop the process
            source_id: Source queue ID
            source_queue: Source queue object
            dest_queues_dict: Dictionary holding a set of destination queues
            timeout: How long to wait for a message
            metrics_queue: Where to place metric snapshots (bounded; when full the
                oldest entry is evicted to make room for the newest)
            metric_interval: How many messages between periodic metric emissions
            name: Name to use in logging
        """
        def _emit_metrics():
            """Emit a metrics snapshot, evicting the oldest entry if the queue is full.

            Newer snapshots are more valuable than older ones — they reflect the most
            recent timing and count data — so we prefer to keep the newest rather than
            silently discard it.
            """
            data = metrics.get_all_data()
            try:
                metrics_queue.put_nowait(data)
            except Full:
                try:
                    metrics_queue.get_nowait()  # evict oldest to make room
                except Empty:
                    pass  # consumer drained concurrently; next put will succeed
                try:
                    metrics_queue.put_nowait(data)
                except Full:
                    pass  # still full after eviction (race); accept the loss

        # Make a helpful logger name
        if name is None:
            logger_name = f'{__name__}.publisher.{source_id}'
        else:
            logger_name = f'{__name__}.publisher.{name}.{source_id}'

        log = logging.getLogger(logger_name)
        log.addHandler(logging.NullHandler())

        # Set up metrics
        counter = 0
        metrics = Metrics()

        latency_metric_id = metrics.add_element(MetricType.TIMING, name='pipe latency')
        metrics.start(latency_metric_id)

        counting_metric_id = metrics.add_element(MetricType.COUNTING, name='movement count')

        while True:
            # Check for stop
            if stop_event.is_set():
                log.info("Stopping due to stop event")
                _emit_metrics()
                return

            try:
                log.debug("Trying to get line from source in pair %s",
                          source_id)
                line = safe_get(source_queue, timeout=timeout, stop_event=stop_event)

                # Distribute the line to all downstream queues
                for dest_id, dest_queue in dest_queues_dict.items():
                    log.info("Writing line from source %s to dest %s",
                             source_id, dest_id)
                    dest_queue.put(line)

                # Mark that we've finished processing the item from the queue
                if hasattr(source_queue, 'task_done'):
                    source_queue.task_done()

                # Mark latency
                counter += 1
                metrics.lap(latency_metric_id)
                metrics.increment(counting_metric_id)

                # Send metric info
                if counter % metric_interval == 0:
                    _emit_metrics()

            except Empty:
                log.debug("No lines to get from source in pair %s", source_id)
                continue

            except EOFError:
                log.debug("Source in pair %s no longer available", source_id)

            except BrokenPipeError:
                log.debug("One pipe in pair %s is no longer available",
                          name if name else source_id)
                return

    def get_queue(self, queue_id: Union[str, int]) -> UNION_SUPPORTED_QUEUES:
        """Retrieve a client's queue reference

        Args:
            queue_id: ID of the queue reference

        Returns:
            Reference to a queue instance
        """
        if queue_id in self.client_queues_source:
            queue_list = self.client_queues_source
        else:
            queue_list = self.client_queues_destination

        return queue_list[text(queue_id)]

    @validate_direction
    def register_queue(self,
                       q: UNION_SUPPORTED_QUEUES,
                       direction: DIRECTION) -> str:
        """Register a queue to this link

        For a new ``FROM`` (source) queue, a publishing process will be created to send
        all additions down to destination queues.

        For a new ``TO`` (destination) queue, all new additions to source queues
        will be added to this queue.

        Returns the numeric ID for the queue reference, which must be used in all
        future interactions.

        Args:
            q: Queue object to register
            direction: DIRECTION.FROM or DIRECTION.TO

        Returns:
            The client's ID for access to this queue
        """
        # Calculate the opposite direction
        op_direction = DIRECTION.TO if direction == DIRECTION.FROM else DIRECTION.FROM

        # Warnings
        # Priority Queues
        if isinstance(q, tuple(PRIORITY_QUEUES)):
            self._log.warning('Entries in a PriorityQueue must have the correct format (tuple or '
                           'simple class with a priority value) or the queue will block.')

        # SimpleQueues
        if direction == DIRECTION.FROM and isinstance(q, tuple(SIMPLE_QUEUES)):
            self._log.warning('Using multiple readers on a SimpleQueue that is a source for a '
                              'QueueLink instance (including other QueueLink instances) can cause '
                              'a deadlock. Queue type: %s.', type(q))

        with self.queues_lock:
            # Get the queue list and opposite queue list
            queue_dict = getattr(self, f'client_queues_{direction}')
            op_queue_dict = getattr(self, f'client_queues_{op_direction}')

            # Make sure we don't accidentally create a loop, or add multiple
            # times
            if q in queue_dict.values():
                raise ValueError("Cannot add this queue again")

            if q in op_queue_dict.values():
                raise ValueError('This queue is in the opposite list. Cannot'
                                 f' add to the {direction} list because it would cause'
                                 ' a circular reference.')

            # Increment the current queue_id
            queue_id = self.last_queue_id + 1
            self.last_queue_id = queue_id

            # Store the queue proxy in the appropriate queue list
            queue_dict[text(queue_id)] = q

            # (Re)create the publishing processes
            # New source:
            #   Just add a new publishing process and send it the list of
            #   destinations.
            if direction == DIRECTION.FROM:
                self._log.debug("Registering source client")
                source_id = queue_id

                # Create and store a new stop event
                stop_event = self.Event()
                self.publisher_stops[text(source_id)] = stop_event

                # Start the publisher
                # This doesn't do anything unless there are some available
                # destinations
                self._start_publisher(source_id)

            # New destination:
            #   Stop all existing publishing processes and restart with updated
            #   destination queue list.
            else:
                self._log.debug("Registering destination client")

                # Stop current processes
                self._stop_publishers()

                # Restart publishers with the updated destinations
                for source_id in self.client_queues_source:
                    self._log.debug("Starting publisher %s", source_id)
                    self._start_publisher(source_id)

        return text(queue_id)

    def read(self, q: UNION_SUPPORTED_QUEUES) -> str:
        """Register a source queue

        Args:
            q: Queue or proxy object to a queue

        Returns:
            The client's ID for access to this queue
        """
        return self.register_queue(q=q, direction=DIRECTION.FROM)

    def write(self, q: UNION_SUPPORTED_QUEUES) -> str:
        """Register a destination queue

        Args:
            q: Queue or proxy object to a queue

        Returns:
            The client's ID for access to this queue
        """
        return self.register_queue(q=q, direction=DIRECTION.TO)

    def _start_publisher(self, source_id: str):
        """Eliminate duplicated Process() call in register_queue

        Start a publisher process to move items from the source queue to
        the destination queues

        Should only be called by register_queue and unregisterQueue (they hold
        a lock for the queue dictionaries while this runs.)
        """
        stop_event = self.publisher_stops[text(source_id)]
        source_queue = self.client_queues_source[text(source_id)]
        dest_queues_dict = self.client_queues_destination
        destination_names = "-".join(dest_queues_dict.keys())

        # Make sure we don't overwrite a running publisher (Process or Thread)
        try:
            if self.client_pair_publishers[text(source_id)].is_alive():
                raise ValueError("Cannot overwrite a running publisher!")

        # If this is new, there won't be an existing value here
        except KeyError:
            pass

        # Make sure there is at least one desination queue
        if len(dest_queues_dict) == 0:
            return

        # Start a publisher process to move items from the source
        # queue to the destination queues
        self._log.info("Starting queue link for source %s to %s",
                       source_id, destination_names)

        # Name the process
        if self.log_name is not None:
            process_name = f'ClientPublisher-{self.log_name}-{source_id}'
        else:
            process_name = f'ClientPublisher-{source_id}'

        # Determine whether to use a thread or process based publisher
        threaded = is_threaded([source_queue, *dest_queues_dict.values()])

        # Decide whether to use a local thread or process-based publisher
        # pylint: disable=invalid-name
        Parallel = Thread if threaded or self.thread_only else self.Process
        # pylint: enable=invalid-name

        # Thread publishers use the threading queue (no feeder-thread race);
        # process publishers use the multiprocessing queue (cross-process safe).
        metrics_q = (self._metrics_queue_thread
                     if (threaded or self.thread_only)
                     else self._get_process_metrics_queue())

        # Start the publisher
        proc = Parallel(target=self._publisher,
                        name=process_name,
                        kwargs={"name": self.name,
                                "stop_event": stop_event,
                                "source_id": source_id,
                                "source_queue": source_queue,
                                "dest_queues_dict": dest_queues_dict,
                                "timeout": self.link_timeout,
                                "metrics_queue": metrics_q})
        proc.daemon = True
        proc.start()
        self.started.set()
        self.client_pair_publishers[text(source_id)] = proc

    def _stop_publisher(self, source_id: str) -> None:
        """Stop a current publisher process

        Args:
            source_id: ID of the source queue
        """
        self._log.debug("Stopping publisher %s", source_id)
        stop_event = self.publisher_stops[text(source_id)]
        stop_event.set()

        # Wait for the process to stop
        try:
            proc_old = self.client_pair_publishers[text(source_id)]

            while True:
                proc_old.join(timeout=1)

                if proc_old.is_alive():
                    self._log.info("Waiting for client %s to stop", source_id)
                else:
                    break

        # If no current publishers are running, a KeyError will be raised
        except KeyError:
            pass

        # Flip the "stop" event back to normal
        # Clear the stop event so it can be reused if this publisher is restarted
        # (e.g., when a new destination queue is registered)
        stop_event.clear()

        # Metrics queues are intentionally NOT replaced here. They persist across
        # publisher restarts so that data emitted by the stopping publisher remains
        # available to get_metrics(). The bounded maxsize=100 on each queue ensures
        # that restarted publishers' put_nowait calls discard safely if the consumer
        # falls behind.

    def _stop_publishers(self) -> None:
        """Stop all current publisher processes"""
        self._log.debug("Stopping any current publishers")
        for source_id in self.client_queues_source:
            self._stop_publisher(source_id)

    @validate_direction
    def unregister_queue(self, queue_id: Union[str, int], direction: DIRECTION) -> str:
        """Detach a queue from this link

        Returns the clientId that was removed

        Args:
            queue_id: ID of the client
            direction: source or destination

        Returns:
            ID of the client queue
        """
        queue_id_txt = text(queue_id)

        with self.queues_lock:
            queue_list = getattr(self, f'client_queues_{direction}')
            if queue_id_txt in queue_list:
                queue_list.pop(queue_id_txt)

            if direction == DIRECTION.FROM:
                source_id = queue_id

                # Stop the source publisher
                # Wait for the process to stop
                self._stop_publisher(source_id)

                # Remove the stop
                self.publisher_stops.pop(queue_id_txt)

            else:
                # Stop current processes
                self._stop_publishers()

                # Restart publishers with the updated destinations
                for source_id in self.client_queues_source:
                    self._start_publisher(source_id)

        return queue_id_txt

    def is_empty(self, queue_id: Union[str, None]=None) -> bool:
        """Checks whether the source or destination queues are empty

        Args:
            queue_id: (optional) ID of the client

        Returns:
            If queue_id is None, returns True ONLY if ALL queues are empty.
            If queue_id is provided, True ONLY if both main queue and specified client queue are
            empty.
        """
        with self.queues_lock:
            if queue_id is not None:
                self._log.debug("Checking if %s is empty", queue_id)
                empty = True

                # If this is a downstream queue, we need to make sure all
                # upstream queues are empty, too
                if queue_id in self.client_queues_destination:
                    self._log.debug("First checking upstream queue(s) of %s", queue_id)

                    for source_id in self.client_queues_source:
                        source_empty = self.get_queue(source_id).empty()
                        self._log.info("Source %s is empty: %s",
                                       source_id, source_empty)
                        empty = empty and source_empty

                queue_empty = self.get_queue(queue_id).empty()
                empty = empty and queue_empty
                self._log.info("%s is empty: %s", queue_id, queue_empty)

            else:
                empty = True

                for client_queue in list(self.client_queues_source.values()):
                    empty = empty and client_queue.empty()

                for client_queue in list(self.client_queues_destination.values()):
                    empty = empty and client_queue.empty()

            self._log.debug("Reporting queue link empty: %s", empty)
            return empty

    def is_alive(self) -> bool:
        """Whether all of the publishers are alive

        Returns:
            False if any publisher has stopped

        Raises:
            ProcessNotStarted if we've never started a publisher
        """
        alive = True

        if not self.started.is_set():
            raise ProcessNotStarted(f'{self.name} has not started')

        for proc in self.client_pair_publishers.values():
            alive = alive and proc.is_alive()

        return alive

    def is_drained(self, queue_id: str=None):
        """Check alive and empty

        Attempts clean semantic response to "is there, or will there be, data
        to read?"

        :returns bool
        """
        drained = True

        # Don't want to check "alive", as these processes run until a parent
        # processes stops them (or is stopped).
        # If we haven't started yet, we are not drained
        if not self.started.is_set():
            return False

        drained = drained and self.is_empty(queue_id=queue_id)

        return drained

    @validate_direction
    def destructive_audit(self, direction: DIRECTION):
        """Print a line from each client Queue from the provided direction

        Args:
            direction (DIRECTION): source or destination

        This is a destructive operation, as it *removes* a line from each Queue

        :raises Empty
        """
        with self.queues_lock:
            queue_list = getattr(self, f'client_queues_{direction}')

            for q_id in list(queue_list):
                try:
                    target_queue = self.get_queue(q_id)
                    self._log.info("queue_id %s: %s",
                                   text(q_id), safe_get(target_queue, block=False))
                except Empty:
                    self._log.info("queue_id %s is empty", text(q_id))

    def get_metrics(self) -> dict:
        """Retrieve accumulated metrics from all publishers.

        Drains both the thread and process metrics queues and merges all
        snapshots by element_id. Because each snapshot is keyed by element_id,
        later snapshots for the same element overwrite earlier ones, so the
        result reflects the most recent data for each metric element. Returns
        an empty dict if no metrics have been emitted yet.

        :returns dict mapping element_id -> metric data dict
        """
        value = {}

        queues = [self._metrics_queue_thread]
        if self._metrics_queue_process is not None:
            queues.append(self._metrics_queue_process)

        for mq in queues:
            try:
                while True:
                    value.update(mq.get_nowait())
            except Empty:
                pass

        return value



# -*- coding: utf-8 -*-
"""Custom pipe-queue adapter to read from a pipe and write *text* content to
thread-safe queues"""
from __future__ import unicode_literals

import codecs
import logging
import multiprocessing  # For comparisons

from typing import get_args

from .queue_handle_adapter_base import MessageCounter
from .queue_handle_adapter_base import _QueueHandleAdapterBase
from .contentwrapper import (
    get_len,
    ContentWrapper,
    WRAP_WHEN,
    conditional_wrap)
from .common import (
    UNION_SUPPORTED_EVENTS,
    UNION_SUPPORTED_LOCKS,
    UNION_SUPPORTED_QUEUES,
    DIRECTION,
    UNION_SUPPORTED_IO_TYPES,
    UNION_SUPPORTED_PATH_TYPES)

def connection_readline(self):
    """A readline method for multiprocessing.connection.Connection objects

    Also requires:
        An Event be applied as stop_event to know when to stop.
        A boolean as "self.trusted" to use recv_bytes vs recv

    conn_obj.readline = connection_wrapper.__get__(conn_obj)
    """
    while True:
        try:
            if self.stop_event.is_set():
                return None

            if self.poll(0.005):
                if self.trusted:
                    received = self.recv()
                else:
                    received = self.recv_bytes()

                return received

        except EOFError:
            return None

        except OSError as e:
            # Bad file descriptor / handle is closed
            if e.errno == 9 or str(e).lower() == 'handle is closed':
                return None

            # Some other OS error
            raise e

def add_methods_to_connections(conn: multiprocessing.connection.Connection,
                               trusted: bool) -> multiprocessing.connection.Connection:
    """Attach a new method and attributes to an existing object.

    Args:
        conn: A multiprocessing connection object.
        trusted: Whether to 'trust' the connection.

    Returns:
        The ``conn`` object with added methods and attributes.
    """
    if isinstance(conn, multiprocessing.connection.Connection):
        # Use the "descriptor protocol" to bind the method to an existing object
        # https://docs.python.org/3.13/howto/descriptor.html
        # https://stackoverflow.com/a/2982
        conn.readline = connection_readline.__get__(conn)
        conn.flush = lambda: None
        conn.trusted = trusted

    return conn


# Works around (https://bryceboe.com/2011/01/28/
# the-python-multiprocessing-queue-and-large-objects/ with large objects)
# by using ContentWrapper to buffer large lines to disk when wrap_when is always or auto
class QueueHandleAdapterReader(_QueueHandleAdapterBase):
    """Custom manager to capture the output of processes and store them in
    one more dedicated thread-safe or process-safe queues.
    """

    def __init__(self,
                 queue: UNION_SUPPORTED_QUEUES,
                 *,  # End of positional arguments
                 handle: UNION_SUPPORTED_IO_TYPES=None,
                 name: str=None,
                 log_name: str=None,
                 start_method: str=None,
                 thread_only: bool=None,
                 trusted: bool=False,
                 wrap_when: WRAP_WHEN=WRAP_WHEN.NEVER,
                 wrap_threshold: int=None):
        """
        Read lines of text from a handle or pipe and write (line by line) into a queue.

        Launches a new thread or process to perform the reading/putting. Prefers a new process,
        but if the provided queue is from `queue` will switch to using a thread.

        Args:
            queue: Queue to write to
            handle: An open handle or pipe to consume from
            name: Optional name for this reader
            log_name: Optional name for this reader in log lines
            start_method: Explicit multiprocessing start method to use
            thread_only: Force the adapter to use a thread rather than process
            trusted: Whether to trust Connection objects; True uses .send/.recv, False
                send_bytes/recv_bytes when reading from multiprocessing.connection.Connections
            wrap_when: When to use a ContentWrapper to encapsulate records
            wrap_threshold: Size limit for a line before it is wrapped in a ContentWrapper;
                only applies when wrap_when is WRAP_WHEN.AUTO
        """
        # A multiprocessing.Pipe (Connection) does not have a readline method
        # This checks the type. If not a Connection instance, it returns unchanged
        handle = add_methods_to_connections(conn=handle, trusted=trusted)

        # Check if we can use the pipe directly
        # IO type that does not have readline, and is not a file path
        # get_args syntax used for Python 3.8-3.12 compatibility
        #   https://stackoverflow.com/a/64643971
        if not hasattr(handle, 'readline') and \
          not isinstance(handle, get_args(UNION_SUPPORTED_PATH_TYPES)):
            original_handle = handle
            handle = codecs.getreader('utf-8')(original_handle)

        # Initialize the parent class
        # pylint: disable=bad-super-call
        # Py3 supports super().__init__; this form is kept for backward compat
        super().__init__(queue=queue,
                                         subclass_name=__name__,
                                         queue_direction=DIRECTION.FROM,
                                         name=name,
                                         handle=handle,
                                         log_name=log_name,
                                         start_method=start_method,
                                         thread_only=thread_only,
                                         trusted=trusted,
                                         wrap_when=wrap_when,
                                         wrap_threshold=wrap_threshold)

    # pylint: disable=arguments-differ
    @staticmethod
    def queue_handle_adapter(*,  # All named parameters are required keyword arguments
                             name: str,
                             handle: UNION_SUPPORTED_IO_TYPES,
                             queue: UNION_SUPPORTED_QUEUES,
                             queue_lock: UNION_SUPPORTED_LOCKS,
                             stop_event: UNION_SUPPORTED_EVENTS,
                             messages_processed: MessageCounter,
                             trusted: bool,
                             wrap_when: WRAP_WHEN,
                             wrap_threshold: int):
        """Copy lines from a given pipe handle into a local threading.Queue

        Runs in a separate process, started by __init__. Closes pipe when done
        reading.

        Args:
            name: Name to use in logging
            handle: Handle/pipe/path to read from
            queue: Queue to write to
            queue_lock: Lock used to indicate a write in progress
            stop_event: Used to determine whether to stop the process
            trusted: Whether to trust Connection objects
            wrap_when: When to use a ContentWrapper
            wrap_threshold: Size limit for a line before it is wrapped in a ContentWrapper
        """
        logger_name = f'{__name__}.queue_handle_adapter.{name}'
        log = logging.getLogger(logger_name)
        log.addHandler(logging.NullHandler())

        # Add readline and flush again; needed for spawn and forkserver start_methods
        handle = add_methods_to_connections(conn=handle, trusted=trusted)

        # Open the handle is it is a path
        # get_args syntax used for Python 3.8-3.12 compatibility
        #   https://stackoverflow.com/a/64643971
        handle_name = None
        if isinstance(handle, get_args(UNION_SUPPORTED_PATH_TYPES)):
            handle_name = handle
            handle = open(handle_name, 'r')  # pylint: disable=consider-using-with

        # Calculate the threshold to use
        wrap_threshold = ContentWrapper.THRESHOLD if wrap_threshold is None else wrap_threshold

        # If its a Connection attach the stop_event so our readline can stop
        if isinstance(handle, multiprocessing.connection.Connection):
            handle.stop_event = stop_event

        log.info('Starting reader process')
        log.info('Queue type: %s', type(queue))
        if handle.closed:
            log.warning('Pipe handle is already closed')

        else:
            # Flush out any potentially waiting content
            handle.flush()

            # https://stackoverflow.com/a/2813530
            while True:
                try:
                    line = handle.readline()
                except (ValueError, EOFError):
                    log.info('Unexpected EOF')
                    break

                # Empty line should signify end of file
                if line is None or line == '' or len(line) == 0:
                    break

                # Wrap in a ContentWrapper
                content = conditional_wrap(line, wrap_when=wrap_when, wrap_threshold=wrap_threshold)

                log.info('Read line, trying to get a lock')
                with queue_lock:
                    log.info('Enqueing line of character length %s, byte length %s',
                             len(content), get_len(content))
                    if isinstance(content, ContentWrapper):  # Wrapped Connections return a CW
                        log.debug('Content is in a ContentWrapper')

                    try:
                        queue.put(content)
                        log.debug('Sent to queue')
                    except EOFError:
                        log.info('Unexpected destination queue EOF')
                        break

                    # Increment counter
                    messages_processed.increment()

                # Check whether we should stop now
                if stop_event.is_set():
                    log.info('Reader asked to stop')
                    break

            # Close the handle if we opened it
            if handle_name:
                handle.close()

            # Clean up references
            # Prevent "UserWarning: ResourceTracker called reentrantly for resource cleanup,
            # which is unsupported. The semaphore object '/<name>' might leak."
            queue_lock = None
            stop_event = None
            messages_processed = None
            handle.stop_event = None
            handle = None

        log.info('Reader sub-process complete')



# -*- coding: utf-8 -*-
"""Class to handle metrics for QueueLink

from time import sleep
from queuelink.metrics import Metrics, MetricType

metrics = Metrics()
id = metrics.add_element(MetricType.TIMING, name='')
metrics.start(id)
sleep(1)
metrics.lap(id)
out = metrics.get_all_data()
out == { id: {'name': None, 'data_point_count': 2, 'mean': float, 'median': float, 'stddev': float} }

metrics = Metrics()
id = metrics.add_element(MetricType.COUNTING, name='')
metrics.increment(id)
out = metrics.get_all_data()
out == { id: {'name': None, 'count': 1} }
"""
from __future__ import unicode_literals
from __future__ import annotations

import statistics
from enum import Enum

from .classtemplate import ClassTemplate
from .common import new_id
from .timer import Timer


class MetricType(Enum):
    """Supported metric types"""
    TIMING = 'timing'
    COUNTING = 'counting'


class BaseMetric(ClassTemplate):  # pylint: disable=too-few-public-methods
    """Base class for QueueLink monitoring metrics"""

    def __init__(self, name: str = None, max_points: int = 100):
        self.name = name
        self.max_points = max_points
        self._initialize_logging_with_log_name(__name__)


class CountMetric(BaseMetric):
    """Monitoring metric for counter-based information"""

    def __init__(self, name: str = None):
        super().__init__(name=name)
        self.count = 0

    def increment(self):
        """Add 1 to the internal counter"""
        self.count += 1

    def to_dict(self) -> dict:
        """Serialize this metric to a plain dict"""
        return {
            'name': self.name,
            'count': self.count,
        }


class TimedMetric(BaseMetric):
    """Monitoring metric for timing information"""

    def __init__(self, name: str = None, interval: float = None):
        super().__init__(name=name)
        self.interval = interval
        self.data_points = []
        self.mean = 0.0
        self.median = 0.0
        self.stddev = 0.0

        # Set at various times
        self.timer = None
        self.start_time_epoch = None
        self.last_lap = None

    def start(self):
        """Start timing"""
        kwargs = {} if self.interval is None else {'interval': self.interval}
        self.timer = Timer(**kwargs)
        self.start_time_epoch = self.timer.start_time
        self.last_lap = 0

    def lap(self):
        """Count when an event occurs"""
        seconds_since_start = self.timer.lap()
        seconds_since_last_lap = seconds_since_start - self.last_lap

        self.last_lap = seconds_since_start
        self._update_values(new_value=seconds_since_last_lap)

    def _update_values(self, new_value):
        """Provide new values to this method. Do not increment data_points separately."""
        if len(self.data_points) == self.max_points:
            removed = self.data_points.pop(0)
            self._log.debug('Max value count reached, removed oldest data point "%s"', removed)

        self.data_points.append(new_value)
        self._update_mean()
        self._update_median()
        self._update_stddev()

    def _update_mean(self):
        """Update the average across the available data points"""
        self.mean = statistics.mean(self.data_points)

    def _update_median(self):
        """Update the median across the available data points"""
        self.median = statistics.median(self.data_points)

    def _update_stddev(self):
        """Update the standard deviation across the available data points"""
        self.stddev = statistics.pstdev(self.data_points)

    def to_dict(self) -> dict:
        """Serialize this metric to a plain dict"""
        return {
            'name': self.name,
            'data_point_count': len(self.data_points),
            'mean': self.mean,
            'median': self.median,
            'stddev': self.stddev,
        }


class Metrics(ClassTemplate):
    """Managing multiple metric instances and retrieving values"""

    def __init__(self,
                 name: str = None,
                 log_name: str = None):
        # Unique ID
        self.id = new_id()

        self.name = name
        self.log_name = log_name
        self._initialize_logging_with_log_name(__name__)

        # Element management
        self.elements = {}

    def add_element(self, metric_type: MetricType, name: str = None) -> str:
        """Add a new metric element

        Args:
            metric_type: MetricType.TIMING or MetricType.COUNTING
            name: Optional human-readable name for this element

        Returns:
            element_id string for all future interactions with this element

        Raises:
            ValueError: if metric_type is not a recognised MetricType
        """
        element_id = new_id()
        while element_id in self.elements:
            element_id = new_id()
            self._log.warning('Element ID collision, retrying: %s', str(element_id))

        if metric_type == MetricType.TIMING:
            m = TimedMetric(name=name)
        elif metric_type == MetricType.COUNTING:
            m = CountMetric(name=name)
        else:
            raise ValueError(f'Unknown metric type "{metric_type}"')

        self.elements[element_id] = {'name': name, 'metrics': m}

        return element_id

    def start(self, element_id: str):
        """Start timing metrics"""
        self.elements[element_id]['metrics'].start()

    def lap(self, element_id: str):
        """Update timing metrics based on a recurring event occurrence"""
        self.elements[element_id]['metrics'].lap()

    def increment(self, element_id: str):
        """Increment a counter metric"""
        self.elements[element_id]['metrics'].increment()

    def get(self, element_id: str) -> BaseMetric:
        """Retrieve a single metric instance"""
        return self.elements[element_id]['metrics']

    def get_data(self, element_id: str) -> dict:
        """Retrieve values related to a metric instance"""
        return self.get(element_id).to_dict()

    def get_all_data(self) -> dict:
        """Retrieve data for all metric instances, keyed by element_id

        Returns:
            dict mapping element_id -> metric data dict
        """
        return {eid: self.get_data(eid) for eid in self.elements}
